""" A module which helps prevent eventual consistency issues on the Datastore. """

# SYSTEM
import logging

# 3RD PARTY
from django.conf import settings
from google.appengine.datastore.datastore_rpc import BaseConnection

# CONSISTENCY
from djangae.contrib.consistency.caches import get_caches


DEFAULT_CONFIG = {
    "cache_on_creation": True,
    "cache_on_modification": False,
    "cache_time": 60, # seconds
    "caches": ["django"],
    "only_cache_matching": [],
}



########################## API ##########################

def improve_queryset_consistency(queryset):
    """ Makes a queryset eventual-consistency-resistant (but not immune to it) by:
        1. Explicitly including PKs of recently-created/-modified objects (if they match the query).
        2. Re-fetching each object by PK to ensure that we get the latest version and exclude
           objects which no longer match the query.
    """
    original = queryset.all()
    recent_pks = get_recent_object_pks_for_model(queryset.model)

    max_existing_pks = BaseConnection.MAX_GET_KEYS - len(recent_pks)
    high_mark = queryset.query.high_mark
    low_mark = queryset.query.low_mark or 0
    if high_mark is None or (high_mark - low_mark > max_existing_pks):
        # Having no limit set or a limit which is too high can cause 2 issues:
        # 1. Potential slowness because we fetch PKs for the whole result.
        # 2. Potential death because we might end up with more than 1000 PKs.
        # We avoid 2 and keep 1 to a minimum by imposing a limit to ensure a total result of <= 1000.
        # Note that this is imperfect because not all of the objects from recent_pks will
        # necessarily match the query, so we might limit more than necessary.
        imposed_limit = max_existing_pks + (queryset.query.low_mark or 0)
        logging.info("Limiting queryset for %s to %d", queryset.model, imposed_limit)
        queryset = queryset.all()[:imposed_limit]

    pks = list(queryset.all().values_list('pk', flat=True)) # this may include recently-created objects
    combined_pks = list(set(pks + recent_pks))
    # By using pk__in we cause the objects to be re-fetched with datastore.Get so we get the
    # up-to-date version of every object.
    # We keep the original filtering as well so that objects which don't match the query are excluded.
    # Keeping the original queryset also retains the ordering.
    return original.filter(pk__in=combined_pks)


def get_recent_objects(queryset):
    """ Get and return a queryset of recently-created/-modified objects which match the given queryset.
        You can append/include/merge this with the results of the original queryset as you wish.
        Note that this may include objects which are also returned by your original queryset.
    """
    return queryset.filter(pk__in=get_recent_object_pks_for_model(queryset.model))


######################## SIGNALS ########################


# See signals.py for registraion

def handle_post_save(sender, instance, created, **kwargs):
    config = get_config(sender)
    if should_cache(instance, created, config):
        add_object_pk_to_caches(instance, config)


def handle_post_delete(sender, instance, **kwargs):
    config = get_config(sender)
    if might_be_cached(sender, config):
        remove_object_pk_from_caches(instance, config)


#########################################################


def get_config(model_class):
    """ Get the config for the given model class. """
    model_identifier = u"%s.%s" % (model_class._meta.app_label, model_class._meta.model_name)
    config = DEFAULT_CONFIG.copy()
    overrides = getattr(settings, "CONSISTENCY_CONFIG", {})
    config.update(overrides.get("defaults", {}))
    config.update(overrides.get("models", {}).get(model_identifier, {}))
    return config


def should_cache(obj, created, config):
    if created:
        if not config["cache_on_creation"]:
            return False
    else:
        if not config["cache_on_modification"]:
            return False
    if not config["only_cache_matching"]:
        return True
    return object_matches_a_check(obj, config["only_cache_matching"])


def might_be_cached(obj, config):
    """ Might the given object be cached? """
    if not (config["cache_on_creation"] or config["cache_on_modification"]):
        return False
    if not config["only_cache_matching"]:
        return True
    return object_matches_a_check(obj, config["only_cache_matching"])


def object_matches_a_check(obj, checks):
    """ Does the object match *any* of the given checks from the "only_cache_matching" list? """
    for check in checks:
        if callable(check):
            if check(obj):
                return True
        else:
            try:
                for field, value in check.iteritems():
                    if not getattr(obj, field) == value:
                        break
                else:
                    return True
            except AttributeError:
                logging.error("Invalid filter for model %s, %s", obj.__class__, check)
                raise
    return False


def get_recent_object_pks_for_model(model_class):
    config = get_config(model_class)
    cache_key = get_model_cache_key(model_class)
    pks = set()
    for cache in get_caches(config["caches"]):
        pks.update(cache.get_pks(model_class, config, cache_key))
    return list(pks)


def add_object_pk_to_caches(obj, config):
    cache_key = get_model_cache_key(obj.__class__)
    for cache in get_caches(config["caches"]):
        cache.add(obj, config, cache_key)


def remove_object_pk_from_caches(obj, config):
    cache_key = get_model_cache_key(obj.__class__)
    for cache in get_caches(config["caches"]):
        cache.remove(obj, config, cache_key)


def get_model_cache_key(model_class):
    return "recently-created-{0}-{1}".format(model_class._meta.app_label, model_class._meta.db_table)



from django.db import models
from djangae.contrib import sleuth
from djangae.test import TestCase, inconsistent_db
from djangae.utils import get_next_available_port
from djangae.db.consistency import ensure_instance_consistent, ensure_instances_consistent


class AvailablePortTests(TestCase):

    def test_get_next_available_port(self):
        url = "127.0.0.1"
        port = 8081
        self.assertEquals(8081, get_next_available_port(url, port))
        with sleuth.switch("djangae.utils.port_is_open",
                lambda *args, **kwargs: False if args[1] < 8085 else True):
            self.assertEquals(8085, get_next_available_port(url, port))


class EnsureCreatedModel(models.Model):
    field1 = models.IntegerField()

    class Meta:
        app_label = "djangae"

    def __unicode__(self):
        return u"PK: {}, field1 {}".format(self.pk, self.field1)


class EnsureCreatedTests(TestCase):

    def test_basic_usage(self):
        for i in xrange(5):
            EnsureCreatedModel.objects.create(
                pk=i + 1,
                field1=i
            )

        with inconsistent_db():
            new_instance = EnsureCreatedModel.objects.create(
                pk=7,
                field1=3
            )

            qs = EnsureCreatedModel.objects.all()
            self.assertEqual(5, len(qs)) # Make sure we don't get the new instance
            self.assertEqual(6, len(ensure_instance_consistent(qs, new_instance.pk))) # Make sure we do
            self.assertEqual(5, len(ensure_instance_consistent(qs[:5], new_instance.pk))) # Make sure slicing returns the right number
            self.assertEqual(3, len(ensure_instance_consistent(qs[2:5], new_instance.pk))) # Make sure slicing returns the right number

            evaled = [ x for x in ensure_instance_consistent(qs.order_by("field1"), new_instance.pk) ]
            self.assertEqual(new_instance, evaled[4]) # Make sure our instance is correctly positioned

            evaled = [ x for x in ensure_instance_consistent(qs.order_by("-field1"), new_instance.pk) ]
            self.assertEqual(new_instance, evaled[1], evaled) # Make sure our instance is correctly positioned

            evaled = [ x for x in ensure_instance_consistent(qs.order_by("-field1"), new_instance.pk)[:2] ]
            self.assertEqual(new_instance, evaled[1], evaled) # Make sure our instance is correctly positioned

            another_instance = EnsureCreatedModel.objects.create(
                pk=8,
                field1=3
            )

            self.assertEqual(5, len(qs)) # Make sure we don't get the new instance
            self.assertEqual(7, len(ensure_instances_consistent(qs, [new_instance.pk, another_instance.pk]))) # Make sure we do

            instance_id = another_instance.pk
            another_instance.delete()

            self.assertEqual(6, len(ensure_instances_consistent(qs, [new_instance.pk, instance_id]))) # Make sure we do

        # Now we're in consistent land!
        self.assertTrue(EnsureCreatedModel.objects.filter(pk=7)[:1])
        self.assertEqual(6, len(ensure_instance_consistent(qs, new_instance.pk)))

        # Make sure it's not returned if it was deleted
        new_instance.delete()
        self.assertEqual(5, len(ensure_instance_consistent(qs, 7)))

        new_instance = EnsureCreatedModel.objects.create(pk=8, field1=8)
        self.assertEqual(1, list(ensure_instance_consistent(qs, 8)).count(new_instance))

    def test_add_many_instances(self):
        for i in xrange(5):
            EnsureCreatedModel.objects.create(
                pk=i + 1,
                field1=i + 5
            )

        with inconsistent_db():
            new_instances = []
            for i in xrange(3):
                instance = EnsureCreatedModel.objects.create(
                    pk=i + 7,
                    field1=i
                )
                new_instances.append(instance)

            new_instance_pks = [i.pk for i in new_instances]

            qs = EnsureCreatedModel.objects.all()
            self.assertEqual(5, len(qs))  # Make sure we don't get the new instance
            self.assertEqual(8, len(ensure_instances_consistent(qs, new_instance_pks)))

            evaled = [ x for x in ensure_instances_consistent(qs.order_by("field1"), new_instance_pks) ]
            self.assertEqual(new_instances[0], evaled[0])  # Make sure our instance is correctly positioned
            self.assertEqual(new_instances[1], evaled[1])  # Make sure our instance is correctly positioned
            self.assertEqual(new_instances[2], evaled[2])  # Make sure our instance is correctly positioned

        # Now we're in consistent land!
        self.assertTrue(EnsureCreatedModel.objects.filter(pk=7)[:1])
        self.assertTrue(EnsureCreatedModel.objects.filter(pk=8)[:1])
        self.assertEqual(8, len(ensure_instances_consistent(qs, new_instance_pks)))

        # Make sure it's not returned if it was deleted
        for instance in new_instances:
            instance.delete()
        self.assertEqual(5, len(ensure_instances_consistent(qs, new_instance_pks)))

    def test_delete_many_instances(self):
        for i in xrange(5):
            EnsureCreatedModel.objects.create(
                pk=i + 1,
                field1=i + 5
            )

        instances_to_delete = []
        for i in xrange(3):
            instance = EnsureCreatedModel.objects.create(
                pk=i + 7,
                field1=i + 1
            )
            instances_to_delete.append(instance)

        instances_to_delete_pks = [i.pk for i in instances_to_delete]

        qs = EnsureCreatedModel.objects.all()
        self.assertEqual(8, len(ensure_instances_consistent(qs, instances_to_delete_pks)))
        with inconsistent_db():
            # Make sure it's not returned if it was deleted
            for instance in instances_to_delete:
                instance.delete()
            self.assertEqual(5, len(ensure_instances_consistent(qs, instances_to_delete_pks)))



import os
import re
from optparse import make_option

from django.core.management.commands.runserver import BaseRunserverCommand
from django.conf import settings

from datetime import datetime

from google.appengine.tools.devappserver2 import shutdown
from google.appengine.tools.sdk_update_checker import (
    GetVersionObject,
    _VersionList
)

DJANGAE_RUNSERVER_IGNORED_FILES_REGEXES = getattr(settings, "DJANGAE_RUNSERVER_IGNORED_FILES_REGEXES", [])
DJANGAE_RUNSERVER_IGNORED_DIR_REGEXES = getattr(settings, "DJANGAE_RUNSERVER_IGNORED_DIR_REGEXES", [])
if DJANGAE_RUNSERVER_IGNORED_FILES_REGEXES:
    DJANGAE_RUNSERVER_IGNORED_FILES_REGEXES = [re.compile(regex) for regex in DJANGAE_RUNSERVER_IGNORED_FILES_REGEXES]
if DJANGAE_RUNSERVER_IGNORED_DIR_REGEXES:
    DJANGAE_RUNSERVER_IGNORED_DIR_REGEXES = [re.compile(regex) for regex in DJANGAE_RUNSERVER_IGNORED_DIR_REGEXES]


def ignore_file(filename):
    """ Replacement for devappserver2.watchter_common.ignore_file
        - to be monkeypatched into place.
    """
    from google.appengine.tools.devappserver2 import watcher_common
    filename = os.path.basename(filename)
    return(
        filename.startswith(watcher_common._IGNORED_PREFIX) or
        any(filename.endswith(suffix) for suffix in watcher_common._IGNORED_FILE_SUFFIXES) or
        watcher_common._IGNORED_REGEX.match(filename) or
        any(regex.match(filename) for regex in DJANGAE_RUNSERVER_IGNORED_FILES_REGEXES)
    )


def skip_ignored_dirs(dirs):
    """ Replacement for devappserver2.watchter_common.skip_ignored_dirs
    - to be monkeypatched into place.
    """
    # Note that this function modifies the `dirs` list in place, it doesn't return anything.
    # Also note that `dirs` is a list of dir *names* not dir *paths*, which means that we can't
    # differentiate between /foo/bar and /moo/bar because we just get 'bar'. But allowing that
    # would require a whole load more monkey patching.
    from google.appengine.tools.devappserver2 import watcher_common
    watcher_common._remove_pred(dirs, lambda d: d.startswith(watcher_common._IGNORED_PREFIX))
    watcher_common._remove_pred(
        dirs,
        lambda d: any(regex.search(d) for regex in DJANGAE_RUNSERVER_IGNORED_DIR_REGEXES)
    )


class Command(BaseRunserverCommand):
    """
    Overrides the default Django runserver command.

    Instead of starting the default Django development server this
    command fires up a copy of the full fledged App Engine
    dev_appserver that emulates the live environment your application
    will be deployed to.
    """
    # We use this list to prevent user using certain dev_appserver options that
    # might collide with some Django settings.
    WHITELISTED_DEV_APPSERVER_OPTIONS = [
        'A',
        'admin_host',
        'admin_port',
        'auth_domain',
        'storage_path',
        'log_level',
        'max_module_instances',
        'use_mtime_file_watcher',
        'appidentity_email_address',
        'appidentity_private_key_path',
        'blobstore_path',
        'datastore_path',
        'clear_datastore',
        'datastore_consistency_policy',
        'require_indexes',
        'auto_id_policy',
        'logs_path',
        'show_mail_body',
        'enable_sendmail',
        'prospective_search_path',
        'clear_prospective_search',
        'search_indexes_path',
        'clear_search_indexes',
        'enable_task_running',
        'allow_skipped_files',
        'api_port',
        'dev_appserver_log_level',
        'skip_sdk_update_check',
        'default_gcs_bucket_name',
    ]

    def __new__(cls, *args, **kwargs):
        # We need to dynamically populate the `option_list` attribute
        # in order to Django allow extra parameters that we're going to pass
        # to GAE's `dev_appserver.py`
        instance = BaseRunserverCommand.__new__(cls, *args, **kwargs)
        sandbox_options = cls._get_sandbox_options()
        for option in sandbox_options:
            if option in cls.WHITELISTED_DEV_APPSERVER_OPTIONS:
                instance.option_list += (
                    make_option('--%s' % option, action='store', dest=option),
                )
        return instance

    @classmethod
    def _get_sandbox_options(cls):
        # We read the options from Djangae's sandbox
        from djangae import sandbox
        return [option for option in dir(sandbox._OPTIONS) if not option.startswith('_')]

    def handle(self, addrport='', *args, **options):
        from djangae import sandbox

        self.gae_options = {}
        sandbox_options = self._get_sandbox_options()

        # this way we populate the dictionary with the options that relevant
        # just for `dev_appserver.py`
        for option, value in options.items():
            if option in sandbox_options and value is not None:
                self.gae_options[option] = value

        super(Command, self).handle(addrport=addrport, *args, **options)

    def run(self, *args, **options):
        self.use_reloader = options.get("use_reloader")
        options["use_reloader"] = False
        return super(Command, self).run(*args, **options)

    def inner_run(self, *args, **options):
        import sys

        shutdown_message = options.get('shutdown_message', '')

        quit_command = 'CTRL-BREAK' if sys.platform == 'win32' else 'CONTROL-C'

        from djangae.environment import get_application_root
        from djangae.sandbox import _find_sdk_from_python_path
        from djangae.blobstore_service import stop_blobstore_service

        from django.conf import settings
        from django.utils import translation

        stop_blobstore_service()

        # Check for app.yaml
        expected_path = os.path.join(get_application_root(), "app.yaml")
        if not os.path.exists(expected_path):
            sys.stderr.write("Unable to find app.yaml at '%s'\n" % expected_path)
            sys.exit(1)

        self.stdout.write("Validating models...\n\n")
        self.check(display_num_errors=True)
        self.stdout.write((
            "%(started_at)s\n"
            "Django version %(version)s, using settings %(settings)r\n"
            "Starting development server at http://%(addr)s:%(port)s/\n"
            "Quit the server with %(quit_command)s.\n"
        ) % {
            "started_at": datetime.now().strftime('%B %d, %Y - %X'),
            "version": self.get_version(),
            "settings": settings.SETTINGS_MODULE,
            "addr": self._raw_ipv6 and '[%s]' % self.addr or self.addr,
            "port": self.port,
            "quit_command": quit_command,
        })
        sys.stdout.write("\n")
        sys.stdout.flush()

        # django.core.management.base forces the locale to en-us. We should
        # set it up correctly for the first request (particularly important
        # in the "--noreload" case).
        translation.activate(settings.LANGUAGE_CODE)

        # Will have been set by setup_paths
        sdk_path = _find_sdk_from_python_path()

        from google.appengine.tools.devappserver2 import devappserver2
        from google.appengine.tools.devappserver2 import python_runtime

        from djangae import sandbox

        # Add any additional modules specified in the settings
        additional_modules = getattr(settings, "DJANGAE_ADDITIONAL_MODULES", [])
        if additional_modules:
            sandbox._OPTIONS.config_paths.extend(additional_modules)

        if int(self.port) != sandbox._OPTIONS.port or additional_modules:
            # Override the port numbers
            sandbox._OPTIONS.port = int(self.port)
            sandbox._OPTIONS.admin_port = int(self.port) + len(additional_modules) + 1
            sandbox._OPTIONS.api_port = int(self.port) + len(additional_modules) + 2

        if self.addr != sandbox._OPTIONS.host:
            sandbox._OPTIONS.host = sandbox._OPTIONS.admin_host = sandbox._OPTIONS.api_host = self.addr

        # Extra options for `dev_appserver.py`
        for param, value in self.gae_options.items():
            setattr(sandbox._OPTIONS, param, value)

        # External port is a new flag introduced in 1.9.19
        current_version = _VersionList(GetVersionObject()['release'])
        if current_version >= _VersionList('1.9.19'):
            sandbox._OPTIONS.external_port = None

        sandbox._OPTIONS.automatic_restart = self.use_reloader

        if sandbox._OPTIONS.host == "127.0.0.1" and os.environ["HTTP_HOST"].startswith("localhost"):
            hostname = "localhost"
            sandbox._OPTIONS.host = "localhost"
        else:
            hostname = sandbox._OPTIONS.host

        os.environ["HTTP_HOST"] = "%s:%s" % (hostname, sandbox._OPTIONS.port)
        os.environ['SERVER_NAME'] = os.environ['HTTP_HOST'].split(':', 1)[0]
        os.environ['SERVER_PORT'] = os.environ['HTTP_HOST'].split(':', 1)[1]
        os.environ['DEFAULT_VERSION_HOSTNAME'] = '%s:%s' % (os.environ['SERVER_NAME'], os.environ['SERVER_PORT'])

        from google.appengine.api.appinfo import EnvironmentVariables

        class NoConfigDevServer(devappserver2.DevelopmentServer):
            def _create_api_server(self, request_data, storage_path, options, configuration):
                # sandbox._create_dispatcher returns a singleton dispatcher instance made in sandbox
                self._dispatcher = sandbox._create_dispatcher(configuration, options)
                # the dispatcher may have passed environment variables, it should be propagated
                env_vars = self._dispatcher._configuration.modules[0]._app_info_external.env_variables or EnvironmentVariables()
                for module in configuration.modules:
                    module_name = module._module_name
                    if module_name == 'default' or module_name is None:
                        module_settings = 'DJANGO_SETTINGS_MODULE'
                    else:
                        module_settings = '%s_DJANGO_SETTINGS_MODULE' % module_name
                    if module_settings in env_vars:
                        module_env_vars = module.env_variables or EnvironmentVariables()
                        module_env_vars['DJANGO_SETTINGS_MODULE'] = env_vars[module_settings]

                        old_env_vars = module._app_info_external.env_variables
                        new_env_vars = EnvironmentVariables.Merge(module_env_vars, old_env_vars)
                        module._app_info_external.env_variables = new_env_vars
                self._dispatcher._configuration = configuration
                self._dispatcher._port = options.port
                self._dispatcher._host = options.host

                self._dispatcher.request_data = request_data
                request_data._dispatcher = self._dispatcher

                sandbox._API_SERVER._host = options.api_host
                sandbox._API_SERVER.bind_addr = (options.api_host, options.api_port)

                from google.appengine.api import apiproxy_stub_map
                task_queue = apiproxy_stub_map.apiproxy.GetStub('taskqueue')
                # Make sure task running is enabled (it's disabled in the sandbox by default)
                if not task_queue._auto_task_running:
                    task_queue._auto_task_running = True
                    task_queue.StartBackgroundExecution()

                return sandbox._API_SERVER

        from google.appengine.tools.devappserver2 import module

        def fix_watcher_files(regex):
            """ Monkeypatch dev_appserver's file watcher to ignore any unwanted dirs or files. """
            from google.appengine.tools.devappserver2 import watcher_common
            watcher_common._IGNORED_REGEX = regex
            watcher_common.ignore_file = ignore_file
            watcher_common.skip_ignored_dirs = skip_ignored_dirs

        regex = sandbox._CONFIG.modules[0].skip_files
        if regex:
            fix_watcher_files(regex)

        def logging_wrapper(func):
            """
                Changes logging to use the DJANGO_COLORS settings
            """
            def _wrapper(level, format, *args, **kwargs):
                if args and len(args) == 1 and isinstance(args[0], dict):
                    args = args[0]
                    status = str(args.get("status", 200))
                else:
                    status = "200"

                try:
                    msg = format % args
                except UnicodeDecodeError:
                    msg += "\n" # This is what Django does in WSGIRequestHandler.log_message

                # Utilize terminal colors, if available
                if status[0] == '2':
                    # Put 2XX first, since it should be the common case
                    msg = self.style.HTTP_SUCCESS(msg)
                elif status[0] == '1':
                    msg = self.style.HTTP_INFO(msg)
                elif status == '304':
                    msg = self.style.HTTP_NOT_MODIFIED(msg)
                elif status[0] == '3':
                    msg = self.style.HTTP_REDIRECT(msg)
                elif status == '404':
                    msg = self.style.HTTP_NOT_FOUND(msg)
                elif status[0] == '4':
                    # 0x16 = Handshake, 0x03 = SSL 3.0 or TLS 1.x
                    if status.startswith(str('\x16\x03')):
                        msg = ("You're accessing the development server over HTTPS, "
                            "but it only supports HTTP.\n")
                    msg = self.style.HTTP_BAD_REQUEST(msg)
                else:
                    # Any 5XX, or any other response
                    msg = self.style.HTTP_SERVER_ERROR(msg)

                return func(level, msg)
            return _wrapper

        module.logging.log = logging_wrapper(module.logging.log)

        python_runtime._RUNTIME_PATH = os.path.join(sdk_path, '_python_runtime.py')
        python_runtime._RUNTIME_ARGS = [sys.executable, python_runtime._RUNTIME_PATH]

        dev_server = NoConfigDevServer()

        try:
            dev_server.start(sandbox._OPTIONS)
            try:
                shutdown.wait_until_shutdown()
            except KeyboardInterrupt:
                pass
        finally:
            dev_server.stop()


        if shutdown_message:
            sys.stdout.write(shutdown_message)

        return



from django.conf.urls import url
from djangae.utils import djangae_webapp

from django.views.decorators.csrf import csrf_exempt

try:
    from mapreduce.main import create_handlers_map
    wrapped_urls = [url(url_re.replace('.*/', '^', 1), csrf_exempt(djangae_webapp(func))) for url_re, func in create_handlers_map()]
except ImportError as e:
    wrapped_urls = []


urlpatterns = wrapped_urls



#STANDARD LIB
from urlparse import urlparse

# LIBRARIES
from django.contrib.auth import get_user_model, get_user, BACKEND_SESSION_KEY
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import HttpRequest
from django.test import TestCase
from django.test.utils import override_settings
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth.hashers import make_password
from google.appengine.api import users

# DJANGAE
from djangae.contrib.gauth.datastore.models import GaeDatastoreUser, Group, get_permission_choices
from djangae.contrib.gauth.datastore.backends import AppEngineUserAPIBackend
from djangae.contrib.gauth.middleware import AuthenticationMiddleware
from djangae.contrib.gauth.settings import AUTHENTICATION_BACKENDS
from djangae.contrib.gauth.utils import get_switch_accounts_url
from djangae.contrib import sleuth


class BackendTests(TestCase):
    """ Tests for the AppEngineUserAPIBackend auth backend. """

    def test_invalid_credentials_cause_typeerror(self):
        """ If the `authenticate` method is passed credentials which it doesn't understand then
            Django expects it to raise a TypeError.
        """
        backend = AppEngineUserAPIBackend()
        credentials = {'username': 'ted', 'password': 'secret'}
        self.assertRaises(TypeError, backend.authenticate, **credentials)

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_authenticate_creates_user_object(self):
        """ If `authenticate` is called with valid credentials then a User object should be created
        """
        User = get_user_model()
        self.assertEqual(User.objects.count(), 0)
        email = 'UpperCasedAddress@example.com'
        google_user = users.User(email, _user_id='111111111100000000001')
        backend = AppEngineUserAPIBackend()
        user = backend.authenticate(google_user=google_user,)
        self.assertEqual(user.email, email.lower())
        self.assertEqual(User.objects.count(), 1)
        # Calling authenticate again with the same credentials should not create another user
        user2 = backend.authenticate(google_user=google_user)
        self.assertEqual(user.pk, user2.pk)

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_user_pre_creation(self):
        """ User objects for Google-Accounts-based users should be able to be pre-created in DB and
            then matched by email address when they log in.
        """
        User = get_user_model()
        backend = AppEngineUserAPIBackend()
        email = '1@example.com'
        # Pre-create our user
        User.objects.pre_create_google_user(email)
        # Now authenticate this user via the Google Accounts API
        google_user = users.User(email=email, _user_id='111111111100000000001')
        user = backend.authenticate(google_user=google_user)
        # Check things
        self.assertEqual(user.email, email)
        self.assertIsNotNone(user.last_login)
        self.assertFalse(user.has_usable_password())

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_user_id_switch(self):
        """ Users sometimes login with the same email, but a different google user id. We handle those cases by
            blanking out the email on the old user object and creating a new one with the new user id.
        """
        email = 'user@customexample.com'
        old_user = users.User(email=email, _user_id='111111111100000000001')
        new_user = users.User(email=email, _user_id='111111111100000000002')
        User = get_user_model()
        backend = AppEngineUserAPIBackend()
        # Authenticate 1st time, creating the user
        user1 = backend.authenticate(google_user=old_user)
        self.assertEqual(user1.email, email)
        self.assertTrue(user1.username.endswith('1'))
        self.assertEqual(1, User.objects.count())

        # Now another user logs in using the same email
        user2 = backend.authenticate(google_user=new_user)
        self.assertEqual(user2.email, email)
        self.assertTrue(user2.username.endswith('2'))
        self.assertEqual(2, User.objects.count())

        # The old account is kept around, but the email is blanked
        user1 = User.objects.get(pk=user1.pk)
        self.assertEqual(user1.email, None)

    @override_settings(DJANGAE_FORCE_USER_PRE_CREATION=True)
    def test_force_user_pre_creation(self):
        User = get_user_model()
        self.assertEqual(User.objects.count(), 0)
        google_user = users.User('1@example.com', _user_id='111111111100000000001')
        backend = AppEngineUserAPIBackend()

        self.assertIsNone(backend.authenticate(google_user=google_user,))
        self.assertEqual(User.objects.count(), 0)

        # superusers don't need pre-creation of User object.
        self.assertEqual(User.objects.count(), 0)
        with sleuth.switch('google.appengine.api.users.is_current_user_admin', lambda: True):
            user = backend.authenticate(google_user=google_user,)
        self.assertEqual(User.objects.count(), 1)
        self.assertEquals(User.objects.get(), user)

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_user_creation_race_condition(self):
        """ If a user double clicks a 'login' button or something, causing 2 threads to be
            authenticating the same user at the same time, ensure it doesn't die.
        """
        email = "test@example.com"
        user_id = "111111111100000000001"
        original_user_get = get_user_model().objects.get

        def crazy_user_get_patch(**kwargs):
            """ Patch for User.objects.get which simulates another thread creating the same user
                immedidately after this is called (by doing it as part of this function). """
            User = get_user_model()
            try:
                return original_user_get(**kwargs) # We patched .get()
            except User.DoesNotExist:
                # This is horrible, but... the backend first tries get() by username and then tries
                # get() by email, and we only want to create our user after that second call
                if kwargs.keys() == ['email']:
                    User.objects.create_user(username=user_id, email=email)
                raise

        backend = AppEngineUserAPIBackend()
        google_user = users.User(email, _user_id=user_id)
        user_class_path = "djangae.contrib.gauth.datastore.models.GaeDatastoreUser.objects.get"
        with sleuth.switch(user_class_path, crazy_user_get_patch):
            backend.authenticate(google_user)


@override_settings(AUTHENTICATION_BACKENDS=AUTHENTICATION_BACKENDS)
class MiddlewareTests(TestCase):
    """ Tests for the AuthenticationMiddleware. """

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_login(self):

        def _get_current_user():
            return users.User('1@example.com', _user_id='111111111100000000001')

        request = HttpRequest()
        SessionMiddleware().process_request(request) # Make the damn sessions work
        request.session[BACKEND_SESSION_KEY] = 'djangae.contrib.gauth.datastore.backends.AppEngineUserAPIBackend'
        middleware = AuthenticationMiddleware()
        # Check that we're not logged in already
        user = get_user(request)
        self.assertFalse(user.is_authenticated())

        # Check that running the middleware when the Google users API doesn't know the current
        # user still leaves us as an anonymous users.
        with sleuth.switch('djangae.contrib.gauth.middleware.users.get_current_user', lambda: None):
            middleware.process_request(request)

        # Check that the middleware successfully logged us in
        user = get_user(request)
        self.assertFalse(user.is_authenticated())

        # Now check that when the Google users API *does* know who we are, that we are logged in.
        with sleuth.switch('djangae.contrib.gauth.middleware.users.get_current_user', _get_current_user):
            middleware.process_request(request)

        # Check that the middleware successfully logged us in
        user = get_user(request)
        self.assertTrue(user.is_authenticated())
        self.assertEqual(user.email, '1@example.com')
        self.assertEqual(user.username, '111111111100000000001')

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_account_switch(self):
        user1 = users.User('1@example.com', _user_id='111111111100000000001')
        user2 = users.User('2@example.com', _user_id='222222222200000000002')

        request = HttpRequest()
        SessionMiddleware().process_request(request)  # Make the damn sessions work
        request.session[BACKEND_SESSION_KEY] = 'djangae.contrib.gauth.datastore.backends.AppEngineUserAPIBackend'
        middleware = AuthenticationMiddleware()

        with sleuth.switch('djangae.contrib.gauth.middleware.users.get_current_user', lambda: user1):
            middleware.process_request(request)

        self.assertEqual(user1.user_id(), request.user.username)

        with sleuth.switch('djangae.contrib.gauth.middleware.users.get_current_user', lambda: user2):
            middleware.process_request(request)

        self.assertEqual(user2.user_id(), request.user.username)

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_user_id_switch(self):
        """ Users sometimes login with the same email, but a different google user id. We handle those cases by
            blanking out the email on the old user object and creating a new one with the new user id.
        """
        email = 'User@example.com'
        user1 = users.User(email, _user_id='111111111100000000001')
        user2 = users.User(email, _user_id='222222222200000000002')

        User = get_user_model()
        request = HttpRequest()
        SessionMiddleware().process_request(request)  # Make the damn sessions work
        request.session[BACKEND_SESSION_KEY] = 'djangae.contrib.gauth.datastore.backends.AppEngineUserAPIBackend'
        middleware = AuthenticationMiddleware()

        with sleuth.switch('djangae.contrib.gauth.middleware.users.get_current_user', lambda: user1):
            middleware.process_request(request)

        self.assertEqual(1, User.objects.count())
        django_user1 = request.user
        self.assertEqual(user1.user_id(), django_user1.username)
        self.assertEqual(user1.email(), django_user1.email)

        with sleuth.switch('djangae.contrib.gauth.middleware.users.get_current_user', lambda: user2):
            middleware.process_request(request)

        self.assertEqual(2, User.objects.count())
        django_user2 = request.user
        self.assertEqual(user2.user_id(), django_user2.username)
        self.assertEqual(user2.email(), django_user2.email)

        django_user1 = User.objects.get(pk=django_user1.pk)
        self.assertEqual(django_user1.email, None)

    @override_settings(DJANGAE_FORCE_USER_PRE_CREATION=True)
    def test_force_user_pre_creation(self):
        email = 'User@example.com'
        user1 = users.User(email, _user_id='111111111100000000001')
        with sleuth.switch('djangae.contrib.gauth.middleware.users.get_current_user', lambda: user1):
            request = HttpRequest()
            SessionMiddleware().process_request(request)  # Make the damn sessions work
            middleware = AuthenticationMiddleware()
            middleware.process_request(request)

        # We expect request.user to be AnonymousUser(), because there was no User object in the DB
        # and so with pre-creation required, authentication should have failed
        self.assertTrue(isinstance(request.user, AnonymousUser))

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_middleware_resaves_email(self):
        # Create user with uppercased email
        email = 'User@example.com'
        google_user = users.User(email, _user_id='111111111100000000001')
        backend = AppEngineUserAPIBackend()
        user = backend.authenticate(google_user=google_user,)
        # Normalize_email should save a user with lowercase email
        self.assertEqual(user.email, email.lower())

        # Run AuthenticationMiddleware, if email are mismatched
        with sleuth.switch('djangae.contrib.gauth.middleware.users.get_current_user', lambda: google_user):
            request = HttpRequest()
            SessionMiddleware().process_request(request)  # Make the damn sessions work
            middleware = AuthenticationMiddleware()
            middleware.process_request(request)

        # Middleware should resave to uppercased email, keeping user the same
        self.assertEqual(request.user.email, email)
        self.assertEqual(request.user.pk, user.pk)


@override_settings(
    AUTH_USER_MODEL='djangae.GaeDatastoreUser',
    AUTHENTICATION_BACKENDS=('djangae.contrib.gauth.datastore.backends.AppEngineUserAPIBackend',)
)
class CustomPermissionsUserModelBackendTest(TestCase):
    """
    Tests for the ModelBackend using the CustomPermissionsUser model.

    As with the ExtensionUser test, this isn't a perfect test, because both
    the User and CustomPermissionsUser are synchronized to the database,
    which wouldn't ordinary happen in production.
    """
    UserModel = GaeDatastoreUser

    def setUp(self):
        # Fix Django so that we can use our custom user model.
        # TODO: Submit a fix to Django to allow override_settings(AUTH_USER_MODEL='something') to
        # work, even if the project has already set AUTH_USER_MODEL to a custom user
        super(CustomPermissionsUserModelBackendTest, self).setUp()

        GaeDatastoreUser.objects = GaeDatastoreUser._default_manager
        GaeDatastoreUser._base_manager = GaeDatastoreUser._default_manager
        self.user = GaeDatastoreUser.objects.create(
            username='test1',
            email='test@example.com',
            password=make_password(None),
            is_active=True,
        )
        self.superuser = GaeDatastoreUser.objects.create(
            username='test2',
            email='test2@example.com',
            is_superuser=True,
            password=make_password(None),
            is_active=True,
        )

    def tearDown(self):
        GaeDatastoreUser.objects.all().delete()
        super(CustomPermissionsUserModelBackendTest, self).tearDown()

    def test_has_perm(self):
        user = self.UserModel.objects.get(pk=self.user.pk)
        self.assertEqual(user.has_perm('auth.test'), False)
        user.is_staff = True
        user.save()
        self.assertEqual(user.has_perm('auth.test'), False)
        user.is_superuser = True
        user.save()
        self.assertEqual(user.has_perm('auth.test'), True)
        user.is_staff = False
        user.is_superuser = False
        user.save()
        self.assertEqual(user.has_perm('auth.test'), False)
        user.is_staff = True
        user.is_superuser = True
        user.is_active = False
        user.save()
        self.assertEqual(user.has_perm('auth.test'), False)

    def test_custom_perms(self):
        user = self.UserModel.objects.get(pk=self.user.pk)
        user.user_permissions = ['auth.test']
        user.save()

        # reloading user to purge the _perm_cache
        user = self.UserModel.objects.get(pk=self.user.pk)
        self.assertEqual(user.get_all_permissions() == set(['auth.test']), True)
        self.assertEqual(user.get_group_permissions(), set([]))
        self.assertEqual(user.has_module_perms('Group'), False)
        self.assertEqual(user.has_module_perms('auth'), True)
        user.user_permissions.extend(['auth.test2', 'auth.test3'])
        user.save()
        user = self.UserModel.objects.get(pk=self.user.pk)
        self.assertEqual(user.get_all_permissions(), set(['auth.test2', 'auth.test', 'auth.test3']))
        self.assertEqual(user.has_perm('test'), False)
        self.assertEqual(user.has_perm('auth.test'), True)
        self.assertEqual(user.has_perms(['auth.test2', 'auth.test3']), True)

        group = Group.objects.create(name='test_group')
        group.permissions = ['auth.test_group']
        group.save()
        user.groups = [group]
        user.save()

        user = self.UserModel.objects.get(pk=self.user.pk)
        exp = set(['auth.test2', 'auth.test', 'auth.test3', 'auth.test_group'])
        self.assertEqual(user.get_all_permissions(), exp)
        self.assertEqual(user.get_group_permissions(), set(['auth.test_group']))
        self.assertEqual(user.has_perms(['auth.test3', 'auth.test_group']), True)

        user = AnonymousUser()
        self.assertEqual(user.has_perm('test'), False)
        self.assertEqual(user.has_perms(['auth.test2', 'auth.test3']), False)

    def test_has_no_object_perm(self):
        """Regressiontest for #12462"""
        user = self.UserModel.objects.get(pk=self.user.pk)
        user.user_permissions = ['auth.test']
        user.save()

        self.assertEqual(user.has_perm('auth.test', 'object'), False)
        self.assertEqual(user.get_all_permissions('object'), set([]))
        self.assertEqual(user.has_perm('auth.test'), True)
        self.assertEqual(user.get_all_permissions(), set(['auth.test']))

    def test_get_all_superuser_permissions(self):
        """A superuser has all permissions. Refs #14795."""
        user = self.UserModel.objects.get(pk=self.superuser.pk)
        self.assertEqual(len(user.get_all_permissions()), len(get_permission_choices()))


@override_settings(
    AUTH_USER_MODEL='djangae.GaeDatastoreUser',
    AUTHENTICATION_BACKENDS=('djangae.contrib.gauth.datastore.backends.AppEngineUserAPIBackend',)
)
class SwitchAccountsTests(TestCase):
    """ Tests for the switch accounts functionality. """

    @override_settings(DJANGAE_CREATE_UNKNOWN_USER=True)
    def test_switch_accounts(self):
        gcu = 'djangae.contrib.gauth.middleware.users.get_current_user'
        final_destination = '/death/' # there's no escaping it
        switch_accounts_url = get_switch_accounts_url(next=final_destination)
        any_url = '/_ah/warmup'
        jekyll = users.User(email='jekyll@gmail.com', _user_id='1')
        hyde = users.User(email='hyde@gmail.com', _user_id='2')

        # we start our scenario with the user logged in
        with sleuth.switch(gcu, lambda: jekyll):
            response = self.client.get(any_url)
            # Check that the user is logged in
            expected_user_query = GaeDatastoreUser.objects.filter(username=jekyll.user_id())
            self.assertEqual(len(expected_user_query), 1)
            self.assertEqual(int(self.client._session()['_auth_user_id']), expected_user_query[0].pk)
            # Now call the switch_accounts view, which should give us a redirect to the login page
            response = self.client.get(switch_accounts_url, follow=False)
            self.assertEqual(response.status_code, 302)
            self.assertEqual(response['location'], users.create_login_url(switch_accounts_url))
            # In tests, we don't have dev_appserver fired up, so we can't actually call the login
            # URL, but let's suppose that the user wasn't logged into multiple accounts at once
            # and so the login page redirected us straight back to the switch_accounts view.
            # It should detect this, and should now redirect us to the log*out* URL with a
            # destination of the log*in* URL
            response = self.client.get(switch_accounts_url)
            self.assertEqual(response.status_code, 302)
            self.assertEqual(
                response['location'],
                users.create_logout_url(users.create_login_url(switch_accounts_url))
            )
            # And now we have to emulate the scenario that we have now logged in with a different
            # account, so re-mock that
        with sleuth.switch(gcu, lambda: hyde):
            # Now that we're logged in as a different user, we expect request.user to get set to
            # the equivalent Django user and to be redirected to our final destination
            response = self.client.get(switch_accounts_url)
            redirect_path = urlparse(response['location']).path # it has the host name as well
            self.assertEqual(redirect_path, final_destination)
            expected_user_query = GaeDatastoreUser.objects.filter(username=hyde.user_id())
            self.assertEqual(len(expected_user_query), 1)
            self.assertEqual(int(self.client._session()['_auth_user_id']), expected_user_query[0].pk)



import copy
import functools
import threading

from google.appengine.api.datastore import (
    CreateTransactionOptions,
    _GetConnection,
    _PushConnection,
    _PopConnection,
    _SetConnection,
    IsInTransaction
)
from google.appengine.datastore.datastore_rpc import TransactionOptions

from djangae.db.backends.appengine import caching


def in_atomic_block():
    # At the moment just a wrapper around App Engine so that
    # users don't have to use two different APIs
    return IsInTransaction()


class ContextDecorator(object):
    """
        A thread-safe ContextDecorator. Subclasses should implement classmethods
        called _do_enter(state, decorator_args) and _do_exit(state, decorator_args, exception)

        state is a thread.local which can store state for each enter/exit. Decorator args holds
        any arguments passed into the decorator or context manager when called.
    """
    VALID_ARGUMENTS = ()

    def __init__(self, func=None, **kwargs):
        # Func will be passed in if this has been called without parenthesis
        # as a @decorator

        # Make sure only valid decorator arguments were passed in
        if len(kwargs) > len(self.__class__.VALID_ARGUMENTS):
            raise ValueError("Unexpected decorator arguments: {}".format(
                set(kwargs.keys()) - set(self.__class__.VALID_ARGUMENTS))
            )

        self.func = func
        self.decorator_args = { x: kwargs.get(x) for x in self.__class__.VALID_ARGUMENTS }
        # Add thread local state for variables that change per-call rather than
        # per insantiation of the decorator
        self.state = threading.local()

    def __get__(self, obj, objtype=None):
        """ Implement descriptor protocol to support instance methods. """
        # Invoked whenever this is accessed as an attribute of *another* object
        # - as it is when wrapping an instance method: `instance.method` will be
        # the ContextDecorator, so this is called.
        # We make sure __call__ is passed the `instance`, which it will pass onto
        # `self.func()`
        return functools.partial(self.__call__, obj)

    def __call__(self, *args, **kwargs):
        # Called if this has been used as a decorator not as a context manager

        def decorated(*_args, **_kwargs):
            decorator_args = self.decorator_args.copy()
            exception = False
            try:
                self.__class__._do_enter(self.state, decorator_args)
                try:
                    return self.func(*_args, **_kwargs)
                except:
                    exception = True
                    raise
            finally:
                self.__class__._do_exit(self.state, decorator_args, exception)

        if not self.func:
            # We were instantiated with args
            self.func = args[0]
            return decorated
        else:
            return decorated(*args, **kwargs)

    def __enter__(self):
        self.__class__._do_enter(self.state, self.decorator_args.copy())

    def __exit__(self, exc_type, exc_value, traceback):
        self.__class__._do_exit(self.state, self.decorator_args.copy(), exc_type)


class TransactionFailedError(Exception):
    pass


class AtomicDecorator(ContextDecorator):
    VALID_ARGUMENTS = ("xg", "independent", "mandatory")

    @classmethod
    def _do_enter(cls, state, decorator_args):
        mandatory = decorator_args.get("mandatory", False)
        independent = decorator_args.get("independent", False)
        xg = decorator_args.get("xg", False)

        # Reset the state
        state.conn_stack = []
        state.transaction_started = False
        state.original_stack = None

        if independent:
            # Unwind the connection stack and store it on the state so that
            # we can replace it on exit
            while in_atomic_block():
                state.conn_stack.append(_PopConnection())
            state.original_stack = copy.deepcopy(caching.get_context().stack)

        elif in_atomic_block():
            # App Engine doesn't support nested transactions, so if there is a nested
            # atomic() call we just don't do anything. This is how RunInTransaction does it
            return
        elif mandatory:
            raise TransactionFailedError("You've specified that an outer transaction is mandatory, but one doesn't exist")

        options = CreateTransactionOptions(
            xg=xg,
            propagation=TransactionOptions.INDEPENDENT if independent else None
        )

        conn = _GetConnection()
        new_conn = conn.new_transaction(options)
        _PushConnection(new_conn)

        assert(_GetConnection())

        # Clear the context cache at the start of a transaction
        caching.get_context().stack.push()
        state.transaction_started = True

    @classmethod
    def _do_exit(cls, state, decorator_args, exception):
        independent = decorator_args.get("independent", False)
        context = caching.get_context()
        try:
            if state.transaction_started:
                if exception:
                    _GetConnection().rollback()
                else:
                    if not _GetConnection().commit():
                        raise TransactionFailedError()
        finally:
            if state.transaction_started:
                _PopConnection()

                 # Clear the context cache at the end of a transaction
                if exception:
                    context.stack.pop(discard=True)
                else:
                    context.stack.pop(apply_staged=True, clear_staged=True)

            # If we were in an independent transaction, put everything back
            # the way it was!
            if independent:
                while state.conn_stack:
                    _PushConnection(state.conn_stack.pop())

                # Restore the in-context cache as it was
                context.stack = state.original_stack


atomic = AtomicDecorator
commit_on_success = AtomicDecorator  # Alias to the old Django name for this kinda thing


class NonAtomicDecorator(ContextDecorator):
    @classmethod
    def _do_enter(cls, state, decorator_args):
        state.conn_stack = []
        context = caching.get_context()

        # We aren't in a transaction, do nothing!
        if not in_atomic_block():
            return

        # Store the current in-context stack
        state.original_stack = copy.deepcopy(context.stack)

        # Similar to independent transactions, unwind the connection statck
        # until we aren't in a transaction
        while in_atomic_block():
            state.conn_stack.append(_PopConnection())

        # Unwind the in-context stack
        while len(context.stack.stack) > 1:
            context.stack.pop(discard=True)

    @classmethod
    def _do_exit(cls, state, decorator_args, exception):
        if not state.conn_stack:
            return

        # Restore the connection stack
        while state.conn_stack:
            _PushConnection(state.conn_stack.pop())

        caching.get_context().stack = state.original_stack


non_atomic = NonAtomicDecorator



# LIBRARIES
from bs4 import BeautifulSoup
from django import forms
from django.db import models

# DJANGAE
from djangae.fields import ListField
from djangae.test import TestCase
from djangae.tests.test_db_fields import JSONFieldModel


class JSONModelForm(forms.ModelForm):
    class Meta:
        model = JSONFieldModel
        fields = ['json_field']


class BlankableListFieldModel(models.Model):
    list_field = ListField(models.CharField(max_length=1), blank=True)


class ListFieldForm(forms.ModelForm):
    class Meta:
        model = BlankableListFieldModel
        fields = ['list_field']


class JSONFieldFormsTest(TestCase):

    def test_json_data_is_python_after_cleaning(self):
        """ In the forms' `cleaned_data`, the json_field data should be python, rather than still
            a string.
        """
        data = dict(json_field="""{"cats": "awesome", "dogs": 46234}""")
        form = JSONModelForm(data)
        assert form.is_valid()  # Sanity, and to trigger cleaned_data
        expected_data = {"cats": "awesome", "dogs": 46234}
        self.assertEqual(form.cleaned_data['json_field'], expected_data)

    def test_json_data_is_rendered_as_json_in_html_form(self):
        """ When the form renders the <textarea> with the JSON in it, it should have been through
            json.dumps, and should not just be repr(python_thing).
        """
        instance = JSONFieldModel(json_field={u'name': 'Lucy', 123: 456})
        form = JSONModelForm(instance=instance)
        html = form.as_p()
        soup = BeautifulSoup(html, "html.parser")
        # Now we want to check that our value was rendered as JSON.
        # So the key 123 should have been converted to a string key of "123"
        textarea = soup.find("textarea").text
        self.assertTrue('"123"' in textarea)


class ListFieldFormsTest(TestCase):

    def test_save_empty_value_from_form(self):
        """ Submitting an empty value in the form should save an empty list. """
        data = dict(list_field="")
        form = ListFieldForm(data)
        self.assertTrue(form.is_valid())
        obj = form.save()
        self.assertEqual(obj.list_field, [])



from djangae.contrib.common import _thread_locals


class RequestStorageMiddleware:
    """ Middleware which allows us to get hold of the request object in places where Django doesn't give it to us, e.g. in model save methods.
        Use get_request() to access the request object.
    """

    def process_request(self, request):
        _thread_locals.request = request

    def process_response(self, request, response):
        # Wipe out the request so that if the following request to this instance doesn't call the middleware (e.g. deferred tasks) we don't end up with randomness
        _thread_locals.request = None
        return response

    def process_exception(self, request, exception):
        _thread_locals.request = None
        return None  # Allow default exception handling to take over



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ImproperlyConfigured


class DjangaeConfig(AppConfig):
    name = 'djangae'
    verbose_name = _("Djangae")

    def ready(self):
        from djangae.db.backends.appengine.caching import reset_context
        from django.core.signals import request_finished, request_started

        request_finished.connect(reset_context, dispatch_uid="request_finished_context_reset")
        request_started.connect(reset_context, dispatch_uid="request_started_context_reset")

        from django.conf import settings
        contenttype_configuration_error = ImproperlyConfigured(
            "If you're using django.contrib.contenttypes, then you need "
            "to add djangae.contrib.contenttypes to INSTALLED_APPS after "
            "django.contrib.contenttypes."
        )
        if 'django.contrib.contenttypes' in settings.INSTALLED_APPS:
            if not 'djangae.contrib.contenttypes' in settings.INSTALLED_APPS:
                # Raise error if User is using Django CT, but not Djangae
                raise contenttype_configuration_error
            else:
                if settings.INSTALLED_APPS.index('django.contrib.contenttypes') > \
                        settings.INSTALLED_APPS.index('djangae.contrib.contenttypes'):
                    # Raise error if User is using both Django and Djangae CT, but
                    # Django CT comes after Djangae CT
                    raise contenttype_configuration_error



import threading
import new
import logging

import django
from django.db import router, connections, models
from django.apps import apps
from django.utils.encoding import smart_text
from django.utils import six
from djangae.crc64 import CRC64


class SimulatedContentTypeManager(models.Manager):
    """
        Simulates content types without actually hitting the datastore.
    """

    _store = threading.local()

    def _get_id(self, app_label, model):
        crc = CRC64()
        crc.append(app_label)
        crc.append(model)
        return crc.fini() - (2 ** 63) #GAE integers are signed so we shift down

    def _get_opts(self, model, for_concrete_model):
        if for_concrete_model:
            model = model._meta.concrete_model
        elif model._deferred:
            model = model._meta.proxy_for_model
        return model._meta

    def _update_queries(self, models):
        """
            This is just to satisfy the contenttypes tests which check that queries are executed at certain
            times. It's a bit hacky but it works for that purpose.
        """
        from django.contrib.contenttypes.models import ContentType
        conn = connections[router.db_for_write(ContentType)]

        if getattr(conn, "use_debug_cursor", getattr(conn, "force_debug_cursor", False)):
            for model in models or []:
                if model not in self._store.queried_models:
                    conn.queries.append("select * from {}".format(ContentType._meta.db_table))
                    break

        self._store.queried_models |= set(models or [])

    def _repopulate_if_necessary(self, models=None):
        if not hasattr(self._store, "queried_models"):
            self._store.queried_models = set()

        if not hasattr(self._store, "constructed_instances"):
            self._store.constructed_instances = {}

        self._update_queries(models)

        if not hasattr(self._store, "content_types"):
            all_models = [ (x._meta.app_label, x._meta.model_name, x) for x in apps.get_models() ]

            self._update_queries([(x[0], x[1]) for x in all_models])

            content_types = {}
            for app_label, model_name, model in all_models:
                content_type_id = self._get_id(app_label, model_name)

                content_types[content_type_id] = {
                    "id": content_type_id,
                    "app_label": app_label,
                    "model": model_name,
                }
                if django.VERSION < (1, 9):
                    content_types[content_type_id]["name"] = smart_text(model._meta.verbose_name_raw)

            self._store.content_types = content_types

    def get_by_natural_key(self, app_label, model):
        self._repopulate_if_necessary(models=[(app_label, model)])
        return self.get(id=self._get_id(app_label, model))

    def get_for_model(self, model, for_concrete_model=True):
        opts = self._get_opts(model, for_concrete_model)
        self._repopulate_if_necessary(models=[(opts.app_label, opts.model_name)])
        return self.get_by_natural_key(opts.app_label, opts.model_name)

    def get_for_models(self, *models, **kwargs):
        for_concrete_model = kwargs.get("for_concrete_models", True)

        self._update_queries(
            [ (self._get_opts(x, for_concrete_model).app_label, self._get_opts(x, for_concrete_model).model_name) for x in models ]
        )
        ret = {}
        for model in models:
            ret[model] = self.get_for_model(model, for_concrete_model)

        return ret

    def get_for_id(self, id):
        return self.get(pk=id)

    def clear_cache(self):
        self._store.queried_models = set()

    def _get_from_store(self, id):
        try:
            return self._store.content_types[id]
        except KeyError:
            from django.contrib.contenttypes.models import ContentType
            raise ContentType.DoesNotExist()

    def get(self, **kwargs):
        from django.contrib.contenttypes.models import ContentType
        self._repopulate_if_necessary()

        if "pk" in kwargs:
            kwargs["id"] = kwargs["pk"]
            del kwargs["pk"]

        if "id" in kwargs:
            dic = self._get_from_store(int(kwargs["id"]))
        else:
            for ct in self._store.content_types.values():
                for k, v in kwargs.items():
                    if k not in ct:
                        raise ContentType.DoesNotExist()

                    if ct[k] != v:
                        break
                else:
                    dic = ct
                    break
            else:
                raise ContentType.DoesNotExist()

        def disable_save(*args, **kwargs):
            raise NotImplementedError("You can't save simulated content types")

        # We do this because some tests to comparisons with 'is' so we store
        # constructed ContentTypes in the thread local and return them if possible
        if dic["id"] in self._store.constructed_instances:
            return self._store.constructed_instances[dic["id"]]
        else:
            from django.contrib.contenttypes.models import ContentType
            result = ContentType(**dic)
            result.save = new.instancemethod(disable_save, ContentType, result)
            self._store.constructed_instances[dic["id"]] = result
            return result

    def create(self, **kwargs):
        from django.contrib.contenttypes.models import ContentType
        try:
            return self.get(**kwargs)
        except ContentType.DoesNotExist:
            logging.warning("Created simulated content type, this will not persist and will remain thread-local")
            new_id = self._get_id(kwargs["app_label"], kwargs["model"])
            kwargs["id"] = new_id
            if "pk" in kwargs:
                del kwargs["pk"]
            self._store.content_types[new_id] = kwargs
            return self.get(id=new_id)

    def get_or_create(self, **kwargs):
        try:
            del kwargs["defaults"]
            return self.get(**kwargs)
        except ContentType.DoesNotExist:
            raise NotImplementedError("You can't manually create simulated content types")

    def filter(self, **kwargs):
        self._repopulate_if_necessary()
        def _condition(ct):
            for attr, val in kwargs.items():
                if getattr(ct, attr) != val:
                    return False
            return True

        return [ct for ct in self.all() if _condition(ct)]

    def all(self, **kwargs):
        result = []

        for ct in self._store.content_types.keys():
            result.append(self.get(id=ct))
        return result

    def using(self, *args, **kwargs):
        return self

    def bulk_create(self, *args, **kwargs):
        pass



import random

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from google.appengine.datastore.datastore_rpc import BaseConnection
from google.appengine.datastore.datastore_stub_util import _MAX_EG_PER_TXN

from djangae.fields.related import (
    RelatedSetField,
    RelatedIteratorManagerBase,
    ReverseRelatedObjectsDescriptor,
)
from djangae.db import transaction


MAX_ENTITIES_PER_GET = BaseConnection.MAX_GET_KEYS

# If the number of shards plus 1 (for the object to which they belong) is <= the max entity groups per
# transaction then we can do the populate() operation in a single transaction, which is nice, hence:
DEFAULT_SHARD_COUNT = MAX_SHARDS_PER_TRANSACTION = _MAX_EG_PER_TXN - 1


class RelatedShardManager(RelatedIteratorManagerBase, models.Manager):
    """ This is what is given to you when you access the field attribute on an instance.  It's a
        model manager with the usual queryset methods (the same as for RelatedSetField) but with
        the additional increment()/decrement()/reset() methods for the counting.
    """

    def increment(self, step=1):
        if step < 0:
            raise ValueError("Tried to increment with a negative number, use decrement instead")

        self._update_or_create_shard(step)

    def decrement(self, step=1):
        if step < 0:
            raise ValueError("Tried to decrement with a negative number, use increment instead")
        self._update_or_create_shard(-step)

    def value(self):
        """ Calcuate the aggregated sum of all the shard values. """
        shards = self.all().values_list('count', flat=True)
        return sum(shards)

    def reset(self):
        """ Reset the counter to 0. """
        # This is not transactional because (1) that wouldn't work with > 24 shards, and (2) if
        # there are other threads doing increments/decrements at the same time then it doesn't make
        # any difference if they happen before or after our increment/decrement anyway.
        value = self.value()
        if value > 0:
            self.decrement(value)
        elif value < 0:
            self.increment(abs(value))

    def clear(self):
        # Override the default `clear` method of the parent class, as that only clears the list of
        # PKs but doesn't delete the related objects.  We want to delete the objects (shards) as well.
        self.reset()

    def populate(self):
        """ Create all the CounterShard objects which will be used by this field. Useful to prevent
            additional saves being caused when you call increment() or decrement() due to having to
            update the list of shard PKs on the instance.
        """
        total_to_create = self.field.shard_count - len(self)
        while total_to_create:
            with transaction.atomic(xg=True):
                # We must re-fetch the instance to ensure that we do this atomically, but we must
                # also update self.instance so that the calling code which is referencing
                # self.instance also gets the updated list of shard PKs
                new_instance = self.instance._default_manager.get(pk=self.instance.pk)
                new_instance_shard_pks = getattr(new_instance, self.field.attname, set())
                # Re-check / update the number to create based on the refreshed instance from the DB
                total_to_create = self.field.shard_count - len(new_instance_shard_pks)
                num_to_create = min(total_to_create, MAX_SHARDS_PER_TRANSACTION)

                new_shard_pks = set()
                for x in xrange(num_to_create):
                    new_shard_pks.add(self._create_shard(count=0).pk)

                new_instance_shard_pks.update(new_shard_pks)
                setattr(self.instance, self.field.attname, new_instance_shard_pks)
                models.Model.save(new_instance) # avoid custom save method, which might do DB lookups
                total_to_create -= num_to_create

    def _update_or_create_shard(self, step):
        """ Find or create a random shard and alter its `count` by the given step. """
        shard_index = random.randint(0, self.field.shard_count - 1)
        # Converting the set to a list introduces some randomness in the ordering, but that's fine
        shard_pks = list(self.field.value_from_object(self.instance)) # needs to be indexable
        try:
            shard_pk = shard_pks[shard_index]
        except IndexError:
            # We don't have this many shards yet, so create a new one
            with transaction.atomic(xg=True):
                # We must re-fetch the instance to ensure that we do this atomically, but we must
                # also update self.instance so that the calling code which is referencing
                # self.instance also gets the updated list of shard PKs
                new_shard = self._create_shard(count=step)
                new_instance = self.instance._default_manager.get(pk=self.instance.pk)
                new_instance_shard_pks = getattr(new_instance, self.field.attname, set())
                new_instance_shard_pks.add(new_shard.pk)
                setattr(self.instance, self.field.attname, new_instance_shard_pks)
                models.Model.save(new_instance) # avoid custom save method, which might do DB lookups
        else:
            with transaction.atomic():
                from djangae.models import CounterShard
                shard = CounterShard.objects.get(pk=shard_pk)
                shard.count += step
                shard.save()

    def _create_shard(self, count):
        from djangae.models import CounterShard
        return CounterShard.objects.create(
            count=count, label="%s.%s" % (self.instance._meta.db_table, self.field.name)
        )


class ReverseRelatedShardsDescriptor(ReverseRelatedObjectsDescriptor):
    """ Subclass of the RelatedSetField's ReverseRelatedObjectsDescriptor which overrides the
        related manager class and prevents setting of the field value directly.
    """
    @property
    def related_manager_cls(self):
        return RelatedShardManager

    def __set__(self, instance, value):
        raise AttributeError(
            "You should not try to set the value of a ShardedCounterField manually. "
            "Use the increment() or decrement() methods."
        )


class ShardedCounterField(RelatedSetField):

    def __init__(self, shard_count=DEFAULT_SHARD_COUNT, **kwargs):
        # Note that by removing the related_name by default we avoid reverse name clashes caused by
        # having multiple ShardedCounterFields on the same model.

        if "default" in kwargs:
            # We don't allow default because it causes all kinds of issues
            raise ImproperlyConfigured(
                "It is not possible to set a default value "
                "on a ShardedCounterField, use increment() after creation instead"
            )

        if "to" in kwargs:
            del kwargs["to"]

        self.shard_count = shard_count
        if shard_count > MAX_ENTITIES_PER_GET:
            raise ImproperlyConfigured(
                "ShardedCounterField.shard_count cannot be more than the Datastore is capable of "
                "fetching in a single Get operation (%d)" % MAX_ENTITIES_PER_GET
            )
        kwargs.setdefault("related_name", "+")
        super(ShardedCounterField, self).__init__('djangae.CounterShard', **kwargs)

    def contribute_to_class(self, cls, name):
        super(ShardedCounterField, self).contribute_to_class(cls, name)
        setattr(cls, self.name, ReverseRelatedShardsDescriptor(self))

    def deconstruct(self):
        name, path, args, kwargs = super(ShardedCounterField, self).deconstruct()

        del kwargs["to"]

        # Add the shard count if necessary
        if self.shard_count != DEFAULT_SHARD_COUNT:
            kwargs["shard_count"] = self.shard_count

        # Default is not a valid kwarg
        if "default" in kwargs:
            del kwargs["default"]

        return name, path, args, kwargs



# STANDARD LIB
from datetime import timedelta
import hashlib
import time

# THRID PARTY
from django.db import models
from django.utils import timezone

# DJANGAE
from djangae.db import transaction
from djangae.fields import CharField


class LockQuerySet(models.query.QuerySet):

    def acquire(self, identifier, wait=True, steal_after_ms=None):
        """ Create or fetch the Lock with the given `identifier`.
        `wait`:
            If True, wait until the Lock is available, otherwise if the lcok is not available then
            return None.
        `steal_after_ms`:
            If the lock is not available (already exists), then steal it if it's older than this.
            E.g. if you know that the section of code you're locking should never take more than
            3 seconds, then set this to 3000.
        """
        identifier_hash = hashlib.md5(identifier).hexdigest()

        @transaction.atomic()
        def trans():
            lock = self.filter(identifier_hash=identifier_hash).first()
            if lock:
                # Lock already exists, so check if it's old enough to ignore/steal
                if (
                    steal_after_ms and
                    timezone.now() - lock.timestamp > timedelta(microseconds=steal_after_ms * 1000)
                ):
                    # We can steal it.  Update timestamp to now and return it
                    lock.timestamp = timezone.now()
                    lock.save()
                    return lock
            else:
                return DatastoreLock.objects.create(
                    identifier_hash=identifier_hash,
                    identifier=identifier
                )

        lock = trans()
        while wait and lock is None:
            time.sleep(0.1)  # Sleep for a bit between retries
            lock = trans()
        return lock


class DatastoreLock(models.Model):
    """ A marker for locking a block of code. """

    objects = LockQuerySet.as_manager()

    # To ensure that different locks are distributed evenly across the Datastore key space, we use
    # a hash of the identifier as the primary key
    identifier_hash = CharField(primary_key=True)
    identifier = CharField()
    timestamp = models.DateTimeField(default=timezone.now)

    def __str__(self):
        return self.identifier

    def release(self):
        self.delete()



from google.appengine.ext.deferred import defer
from google.appengine.runtime import DeadlineExceededError


def _process_shard(model, instance_ids, callback):
    for instance in model.objects.filter(pk__in=instance_ids):
        callback(instance)


def _shard(model, query, callback, shard_size, queue, offset=0):
    keys_queryset = model.objects.all()
    keys_queryset.query = query
    keys_queryset = keys_queryset.values_list("pk", flat=True)

    # Keep iterating until we are done, or we only have 10 seconds to spare!
    while True:
        try:
            ids = list(keys_queryset.all()[offset:offset+shard_size])
            if not ids:
                # We're done!
                return

            # Fire off the first shard
            defer(_process_shard, model, ids, callback, _queue=queue)

            # Increment the offset
            offset += shard_size
        except DeadlineExceededError:
            # If we run out of time, then defer this function again, continuing from the offset
            defer(
                _shard,
                model,
                query,
                callback,
                shard_size,
                queue,
                offset=offset,
                _queue=queue
            )


def defer_iteration(queryset, callback, shard_size=500, _queue="default", _target=None):
    """
        Shards background tasks to call 'callback' with each instance in queryset

        - `queryset` - The queryset to iterate
        - `callback` - A callable which accepts an instance as a parameter
        - `shard_size` - The number instances to process per shard (default 500)
        - `_queue` - The name of the queue to run the shards on

        Note, your callback must be indempotent, shards may retry and your callback
        may be called multiple times on the same instance. If you notice that your
        tasks are receiving DeadlineExceededErrors you probably need to reduce the
        shard size. The shards will work in parallel and will not be sequential.
    """

    # We immediately defer the _shard function so that we don't hold up execution
    defer(_shard, queryset.model, queryset.query, callback, shard_size, _queue, _queue=_queue, _target=_target)



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _
from django.contrib.contenttypes.management import update_contenttypes as django_update_contenttypes
from django.db.models.signals import post_migrate

from .management import update_contenttypes
from .models import SimulatedContentTypeManager


class ContentTypesConfig(AppConfig):
    name = 'djangae.contrib.contenttypes'
    verbose_name = _("Djangae Content Types")
    label = "djangae_contenttypes"

    def ready(self):
        if django_update_contenttypes != update_contenttypes:
            post_migrate.disconnect(django_update_contenttypes)
            from django.contrib.contenttypes import models as django_models
            if not isinstance(django_models.ContentType.objects, SimulatedContentTypeManager):
                django_models.ContentType.objects = SimulatedContentTypeManager()



"""This file exists for backwards compatability.
Please use the abstract or concrete models found in either:
`djangae.contrib.gauth.datastore` for applications using the datastore
backend or `djangae.contrib.gauth.sql` for applications using a relational
database backend
"""

import warnings

from djangae.contrib.gauth.datastore.models import (
    GaeAbstractDatastoreUser, GaeDatastoreUser
)


warnings.warn(
    'GaeAbstractDatastoreUser and GaeDatastoreUser have moved to djangae.contrib.gauth.datastore.models'
)



DEFAULT_FILE_STORAGE = 'djangae.storage.BlobstoreStorage'
FILE_UPLOAD_MAX_MEMORY_SIZE = 1024 * 1024
FILE_UPLOAD_HANDLERS = (
    'djangae.storage.BlobstoreFileUploadHandler',
    'django.core.files.uploadhandler.MemoryFileUploadHandler',
)

DATABASES = {
    'default': {
        'ENGINE': 'djangae.db.backends.appengine'
    }
}

GENERATE_SPECIAL_INDEXES_DURING_TESTING = False

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
    }
}

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        }
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
        'djangae': {
            'level': 'WARN'
        }
    }
}

EMAIL_BACKEND = 'djangae.mail.AsyncEmailBackend'

# Setting to *.appspot.com is OK, because GAE takes care of domain routing
# it needs to be like this because of the syntax of addressing non-default versions
# (e.g. -dot-)
ALLOWED_HOSTS = (".appspot.com", )

DJANGAE_RUNSERVER_IGNORED_FILES_REGEXES = ['^.+$(?<!\.py)(?<!\.yaml)(?<!\.html)']
# Note that these should match a directory name, not directory path:
DJANGAE_RUNSERVER_IGNORED_DIR_REGEXES = [r"^google_appengine$"]



"""
JSONField automatically serializes most Python terms to JSON data.
Creates a TEXT field with a default value of "{}".  See test_json.py for
more information.

 from django.db import models
 from django_extensions.db.fields import json

 class LOL(models.Model):
     extra = json.JSONField()

This field originated from the django_extensions project: https://github.com/django-extensions/django-extensions
"""

from __future__ import absolute_import

import json
from collections import OrderedDict

from django.db import models
from django.conf import settings
from django.utils import six
from django.core.serializers.json import DjangoJSONEncoder

from djangae.forms.fields import JSONFormField, JSONWidget

__all__ = ( 'JSONField',)


def dumps(value):
    return DjangoJSONEncoder().encode(value)


def loads(txt, object_pairs_hook=None):
    value = json.loads(
        txt,
        encoding=settings.DEFAULT_CHARSET,
        object_pairs_hook=object_pairs_hook,
    )
    return value


class JSONDict(dict):
    """
    Hack so repr() called by dumpdata will output JSON instead of
    Python formatted data.  This way fixtures will work!
    """
    def __repr__(self):
        return dumps(self)


class JSONUnicode(six.text_type):
    """
    As above
    """
    def __repr__(self):
        return dumps(self)


class JSONList(list):
    """
    As above
    """
    def __repr__(self):
        return dumps(self)


class JSONOrderedDict(OrderedDict):
    """
    As above
    """
    def __repr__(self):
        return dumps(self)


class JSONField(models.TextField):
    """JSONField is a generic textfield that neatly serializes/unserializes
    JSON objects seamlessly.  Main thingy must be a dict object."""

    def __init__(self, use_ordered_dict=False, *args, **kwargs):
        if 'default' in kwargs:
            if not callable(kwargs['default']):
                raise TypeError("'default' must be a callable (e.g. 'dict' or 'list')")
        else:
            kwargs['default'] = dict

        # use `collections.OrderedDict` rather than built-in `dict`
        self.use_ordered_dict = use_ordered_dict

        models.TextField.__init__(self, *args, **kwargs)

    def parse_json(self, value):
        """Convert our string value to JSON after we load it from the DB"""
        if value is None or value == '':
            return {}
        elif isinstance(value, six.string_types):
            if self.use_ordered_dict:
                res = loads(value, object_pairs_hook=OrderedDict)
            else:
                res = loads(value)
            if isinstance(res, OrderedDict) and self.use_ordered_dict:
                return JSONOrderedDict(res)
            elif isinstance(res, dict):
                return JSONDict(**res)
            elif isinstance(res, six.string_types):
                return JSONUnicode(res)
            elif isinstance(res, list):
                return JSONList(res)
            return res
        else:
            return value

    def to_python(self, value):
        return self.parse_json(value)

    def from_db_value(self, value, expression, connection, context):
        return self.parse_json(value)

    def get_db_prep_save(self, value, connection, **kwargs):
        """Convert our JSON object to a string before we save"""
        if value is None and self.null:
            return None
        return super(JSONField, self).get_db_prep_save(dumps(value), connection=connection)

    def south_field_triple(self):
        """Returns a suitable description of this field for South."""
        # We'll just introspect the _actual_ field.
        from south.modelsinspector import introspector
        field_class = "django.db.models.fields.TextField"
        args, kwargs = introspector(self)
        # That's our definition!
        return (field_class, args, kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(JSONField, self).deconstruct()
        if self.default == {}:
            del kwargs['default']
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        defaults = {
            'form_class': JSONFormField,
            'widget': JSONWidget,
        }
        defaults.update(kwargs)
        return super(JSONField, self).formfield(**defaults)



from django.contrib.auth import authenticate, login, logout, get_user, BACKEND_SESSION_KEY, load_backend
from django.contrib.auth.middleware import AuthenticationMiddleware as DjangoMiddleware
from django.contrib.auth.models import BaseUserManager, AnonymousUser
from djangae.contrib.gauth.common.backends import BaseAppEngineUserAPIBackend

from google.appengine.api import users


class AuthenticationMiddleware(DjangoMiddleware):
    def process_request(self, request):
        django_user = get_user(request)
        google_user = users.get_current_user()

        # Check to see if the user is authenticated with a different backend, if so, just set
        # request.user and bail
        if django_user.is_authenticated():
            backend_str = request.session.get(BACKEND_SESSION_KEY)
            if (not backend_str) or not isinstance(load_backend(backend_str), BaseAppEngineUserAPIBackend):
                request.user = django_user
                return

        if django_user.is_anonymous() and google_user:
            # If there is a google user, but we are anonymous, log in!
            # Note that if DJANGAE_CREATE_UNKNOWN_USER=False then this may not authenticate
            django_user = authenticate(google_user=google_user) or AnonymousUser()
            if django_user.is_authenticated():
                login(request, django_user)

        if django_user.is_authenticated():
            if not google_user:
                # If we are logged in with django, but not longer logged in with Google
                # then log out
                logout(request)
                django_user = AnonymousUser()
            elif django_user.username != google_user.user_id():
                # If the Google user changed, we need to log in with the new one
                logout(request)
                django_user = authenticate(google_user=google_user) or AnonymousUser()
                if django_user.is_authenticated():
                    login(request, django_user)

        # Note that the logic above may have logged us out, hence new `if` statement
        if django_user.is_authenticated():
            # Now make sure we update is_superuser and is_staff appropriately
            is_superuser = users.is_current_user_admin()
            resave = False

            if is_superuser != django_user.is_superuser:
                django_user.is_superuser = django_user.is_staff = is_superuser
                resave = True

            # for users which already exist, we want to verify that their email is still correct
            # users are already authenticated with their user_id, so we can save their real email
            # not the lowercased version
            if django_user.email != google_user.email():
                django_user.email = google_user.email()
                resave = True

            if resave:
                django_user.save()

        request.user = django_user



""" Fake DB API 2.0 for App engine """

class DatabaseError(Exception):
    pass

class IntegrityError(DatabaseError):
    pass

class NotSupportedError(Exception):
    pass

class CouldBeSupportedError(NotSupportedError):
    pass

class DataError(DatabaseError):
    pass

class OperationalError(DatabaseError):
    pass

class InternalError(DatabaseError):
    pass

class ProgrammingError(DatabaseError):
    pass

class InterfaceError(DatabaseError):
    pass

def Binary(val):
    return val

Error = DatabaseError
Warning = DatabaseError



# STANDARD LIB
import logging

# THIRD PARTY
from django.conf import settings
from django.http import HttpResponse
from django.utils import timezone

# DJANGAE
from djangae.contrib.mappers import defer_iteration
from .models import DatastoreLock

# GAE background tasks and crons can run for a maximum of 10 minutes, so in theory you shouldn't
# be locking a block of code which takes longer than that, and even if you're using backends which
# can run for longer, you would still be mad to lock a block of code which runs for > 10 minutes
DELETE_LOCKS_OLDER_THAN_SECONDS = 600
QUEUE = getattr(settings, 'DJANGAE_CLEANUP_LOCKS_QUEUE', 'default')


def cleanup_locks(request):
    """ Delete all Lock objects that are older than 10 minutes. """
    logging.info("Starting djangae.contrib.lock cleanup task")
    cut_off = timezone.now() - timezone.timedelta(seconds=DELETE_LOCKS_OLDER_THAN_SECONDS)
    queryset = DatastoreLock.objects.filter(timestamp__lt=cut_off)
    defer_iteration(queryset, _delete_lock, _queue=QUEUE)
    return HttpResponse("Cleanup locks task is running")


def _delete_lock(lock):
    logging.info("Deleting stale lock '%s' with timestamp %r", lock.identifier, lock.timestamp)
    lock.delete()



# THIRD PARTY
from django.conf.urls import patterns, url

# DJANGAE
from .views import cleanup_locks


urlpatterns = patterns('',
    url(r'^djangae-cleanup-locks/$', cleanup_locks, name="cleanup_locks"),
)



import os
import threading
import logging
import re

blobstore_service = None
server = None

from wsgiref.simple_server import WSGIRequestHandler


class NoLogRequestHandler(WSGIRequestHandler):

    def log_request(self, code='-', size='-'):
        """Normally logs an accepted request. Bug given
        that this is not using global logging but stdout,
        this becomes really annoying in tests. So let's
        not log anything.
        """
        pass


def start_blobstore_service():
    """
        When the blobstore files API was deprecated, the blobstore storage was switched
        to use a POST request to the upload handler when storing files uploaded via Django.

        Unfortunately this breaks in the local sandbox when you aren't running the dev_appserver
        because there is no server to handle the blobstore upload. So, this service is kicked
        off by the local sandbox and only handles blobstore uploads. When runserver kicks in
        this service is stopped.
    """
    global blobstore_service
    global server

    if blobstore_service:
        return

    from wsgiref.simple_server import make_server
    from google.appengine.tools.devappserver2 import blob_upload
    from google.appengine.tools.devappserver2 import blob_image
    from google.appengine.tools.devappserver2 import gcs_server

    from django.core.handlers.wsgi import WSGIRequest
    from django.utils.encoding import force_str
    from socket import error as socket_error

    def call_internal_upload(environ, start_response):
        # Otherwise, just assume it's our internalupload handler
        request = WSGIRequest(environ)

        from djangae.views import internalupload
        response = internalupload(request)

        status = '%s %s' % (response.status_code, response.reason_phrase)
        response_headers = [(str(k), str(v)) for k, v in response.items()]
        start_response(force_str(status), response_headers)
        return response

    def handler(environ, start_response):
        path = environ["PATH_INFO"]

        # If this is an image serving URL, then use use the blob_image WSGI app
        if re.match(blob_image.BLOBIMAGE_URL_PATTERN, path.lstrip("/")):
            return blob_image.Application()(environ, start_response)
        elif re.match(gcs_server.GCS_URL_PATTERN, path.lstrip("/")):
            return gcs_server.Application()(environ, start_response)

        return blob_upload.Application(call_internal_upload)(environ, start_response)

    port = int(os.environ['SERVER_PORT'])
    host = os.environ['SERVER_NAME']
    logging.info("Starting blobstore service on %s:%s", host, port)
    try:
        server = make_server(host, port, handler, handler_class=NoLogRequestHandler)
    except socket_error:
        logging.warning("Not starting blobstore service, it may already be running")
        return

    blobstore_service = threading.Thread(target=server.serve_forever)
    blobstore_service.daemon = True
    blobstore_service.start()


def stop_blobstore_service():
    global blobstore_service
    global server

    if not blobstore_service:
        return

    server.shutdown()
    blobstore_service.join(5)
    blobstore_service = None



#LIBRARIES
import django

from django.db.models.sql import compiler
from django.db.models.expressions import Value, OrderBy
from django.db.models.sql.query import get_order_dir

#DJANGAE
from .commands import (
    SelectCommand,
    InsertCommand,
    UpdateCommand,
    DeleteCommand
)


class SQLCompiler(compiler.SQLCompiler):

    def find_ordering_name(self, name, opts, alias=None, default_order='ASC', already_seen=None):
        """
            Overridden just for the __scatter__ property ordering
        """

        # This allow special appengine properties (e.g. __scatter__) to be supplied as an ordering
        # even though they don't (and can't) exist as Django model fields
        if name.startswith("__") and name.endswith("__"):
            name, order = get_order_dir(name, default_order)
            descending = True if order == 'DESC' else False
            return [ (OrderBy(Value('__scatter__'), descending=descending), False) ]

        return super(SQLCompiler, self).find_ordering_name(
            name,
            opts,
            alias=alias,
            default_order=default_order,
            already_seen=already_seen
        )

    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        self.pre_sql_setup()
        self.refcounts_before = self.query.alias_refcount.copy()

        select = SelectCommand(
            self.connection,
            self.query
        )
        return (select, tuple())

    def get_select(self):
        self.query.select_related = False # Make sure select_related is disabled for all queries
        return super(SQLCompiler, self).get_select()


class SQLInsertCompiler(SQLCompiler, compiler.SQLInsertCompiler):
    def __init__(self, *args, **kwargs):
        self.return_id = None
        super(SQLInsertCompiler, self).__init__(*args, **kwargs)

    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        self.pre_sql_setup()

        from djangae.db.utils import get_concrete_fields

        # Always pass down all the fields on an insert
        return [ (InsertCommand(
            self.connection, self.query.model, self.query.objs,
            list(self.query.fields) + list(get_concrete_fields(self.query.model, ignore_leaf=True)),
            self.query.raw), tuple())
        ]


class SQLDeleteCompiler(SQLCompiler, compiler.SQLDeleteCompiler):
    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        return (DeleteCommand(self.connection, self.query), tuple())


class SQLUpdateCompiler(SQLCompiler, compiler.SQLUpdateCompiler):
    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        self.pre_sql_setup()
        return (UpdateCommand(self.connection, self.query), tuple())


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):
    def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
        if self.query.subquery:
            self.query.high_mark = self.query.subquery.query.high_mark
            self.query.low_mark = self.query.subquery.query.low_mark
        return SQLCompiler.as_sql(self, with_limits, with_col_aliases, subquery)


if django.VERSION < (1, 8):
    from django.db.models.sql.compiler import (
        SQLDateCompiler as DateCompiler,
        SQLDateTimeCompiler as DateTimeCompiler
    )

    class SQLDateCompiler(DateCompiler, SQLCompiler):
        def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
            return SQLCompiler.as_sql(self, with_limits, with_col_aliases, subquery)


    class SQLDateTimeCompiler(DateTimeCompiler, SQLCompiler):
        def as_sql(self, with_limits=True, with_col_aliases=False, subquery=False):
            return SQLCompiler.as_sql(self, with_limits, with_col_aliases, subquery)



from mapreduce import context
from pipeline.util import for_name


def thunk_map(x):
    """
        This is the default map function that wraps the static map function.

        It allows you to pass args and kwargs to your map function for defining
        more dynamic mappers
    """
    ctx = context.get()
    params = ctx.mapreduce_spec.mapper.params
    map_func = params.get('_map', None)
    args = params.get('args', [])
    kwargs = params.get('kwargs', {})
    map_func = for_name(map_func)
    return map_func(x, *args, **kwargs)



import warnings
from collections import deque

from django.db.models.query import QuerySet


def _clone_queryset(qs, klass):
    cloned_qs = klass(
        model=qs.model,
        query=qs.query.clone(),
        using=qs._db,
        hints=qs._hints,
    )
    cloned_qs._for_write = qs._for_write
    cloned_qs._prefetch_related_lookups = qs._prefetch_related_lookups[:]
    cloned_qs._known_related_objects = qs._known_related_objects

    return cloned_qs


def ensure_instances_consistent(queryset, instance_pks):
    """
        HRD fighting generate for iterating querysets.
        In the common case of having just created an item, this
        inserts the item at the correct place in a subsequent query.

        This guarantees that the object will be inserted
        in the right place as per the queryset ordering. Only takes into account
        the query.order_by and not default ordering. Patches welcome!

        If the instance specified by the consistently_got_id was deleted, and it comes back
        in the subsequent query, it is removed from the resultset. This might cause an
        off-by-one number of results if you are slicing the queryset.
    """

    class EnsuredQuerySet(queryset.__class__):
        def _fetch_all(self):
            super(EnsuredQuerySet, self)._fetch_all()

            new_qs = _clone_queryset(qs=self, klass=queryset.__class__)
            new_qs.query.high_mark = None
            new_qs.query.low_mark = 0
            consistently_got = list(new_qs.filter(pk__in=instance_pks))

            def is_less(lhs, rhs):
                for field in queryset.query.order_by:
                    # Go through the fields specified in the order by
                    negative = field.startswith("-")
                    if negative:
                        field = field[1:]
                        # If the ordering was negated, but the lhs < rhs then return false
                        if getattr(lhs, field, None) < getattr(rhs, field, None):
                            return False
                    else:
                        # If the ordering was not negated, and the lhs >= rhs, return false
                        if getattr(lhs, field, None) >= getattr(rhs, field, None):
                            return False
                else:
                    return True

            new_result_cache = []
            count = self.query.high_mark

            # We sort by the is_less function and use a deque so we can pop from the
            # front
            consistently_got = deque(sorted(consistently_got, cmp=is_less))

            # Go through each item in the result set and peek at the first item in the
            # deque. If the item is the same, or the item in the deque should appear before
            # 'item' then add the consistent item. If the consistent item is == item then
            # continue so we don't add a duplicate
            for item in self._result_cache:
                try:
                    consistent_item = consistently_got[0]

                    # Should the consistent item be added?
                    if item == consistent_item or is_less(consistent_item, item):
                        new_result_cache.append(consistent_item)
                        consistently_got.popleft()
                        if item == consistent_item:
                            # Prevent adding a stale duplicate
                            continue
                except IndexError:
                    # No consistent items left, so just append all the things
                    # remaining in the original result set
                    pass

                # Item was deleted, otherwise we would have seen it in consistent_item
                if item.pk in instance_pks:
                    continue

                new_result_cache.append(item)
            else:
                # Finally, if we've been through the entire result set, and we still
                # have consistent things remaining, they must be new and at the end of the
                # ordering so just add them to the end of the result set
                if consistently_got:
                    new_result_cache.extend(list(consistently_got))

            self._result_cache = new_result_cache[:count]

    if isinstance(queryset, QuerySet):
        return _clone_queryset(qs=queryset, klass=EnsuredQuerySet)
    else:
        return queryset

def ensure_instance_consistent(queryset, instance_id):
    return ensure_instances_consistent(queryset, [ instance_id ])

def ensure_instance_included(queryset, instance_id):
    warnings.warn("ensure_instance_included is deprecated, use ensure_instance_consistent instead", DeprecationWarning)
    return ensure_instance_consistent(queryset, instance_id)



import logging

from django.db import models

from djangae.test import TestCase
from djangae.test import process_task_queues
from djangae.contrib.mappers.pipes import MapReduceTask
from djangae.contrib.mappers.defer import defer_iteration, _shard


class TestNode(models.Model):
    data = models.CharField(max_length=32)
    counter = models.IntegerField()


class TestNode2(models.Model):
    data = models.CharField(max_length=32)
    counter = models.IntegerField()


class TestMapperClass(MapReduceTask):

    model = TestNode
    name = 'test_map'

    @staticmethod
    def finish(**kwargs):
        logging.info('TestMapper1 Finished')

    @staticmethod
    def map(entity, *args, **kwargs):
        if entity.counter % 2:
            entity.delete()
            yield ('removed', [entity.pk])
        else:
            yield ('remains', [entity.pk])


class TestMapperClass2(MapReduceTask):

    model = TestNode
    name = 'test_map_2'

    @staticmethod
    def finish(**kwargs):
        logging.info('TestMapper2 Finished')

    @staticmethod
    def map(entity, *args, **kwargs):
        entity.data = "hit"
        entity.save()

class TestMapperClass3(MapReduceTask):

    model = TestNode
    name = 'test_map_3'

    @staticmethod
    def map(entity, *args, **kwargs):
        if all((x in args for x in ['arg1', 'arg2'])) and 'test' in kwargs:
            entity.delete()


class MapReduceTestCase(TestCase):

    def setUp(self):
        super(MapReduceTestCase, self).setUp()

        for x in xrange(10):
            TestNode(data="TestNode".format(x), counter=x).save()

    def test_all_models_delete(self):
        self.assertEqual(TestNode.objects.count(), 10)
        TestMapperClass().start()
        process_task_queues()
        self.assertEqual(TestNode.objects.count(), 5)

    def test_model_init(self):
        """
            Test that overriding the model works
        """
        for x in xrange(10):
            TestNode2(data="TestNode2".format(x), counter=x).save()
        self.assertEqual(TestNode2.objects.count(), 10)
        TestMapperClass(model=TestNode2).start()
        process_task_queues()
        self.assertEqual(TestNode2.objects.count(), 5)

    def test_model_args_kwargs(self):
        """
            Test that overriding the model works
        """
        self.assertEqual(TestNode.objects.count(), 10)
        TestMapperClass3(model=TestNode).start('arg1', 'arg2', test='yes')
        process_task_queues()
        self.assertEqual(TestNode.objects.count(), 0)

    def test_map_fruit_update(self):
        self.assertEqual(TestNode.objects.count(), 10)
        TestMapperClass2().start()
        process_task_queues()
        nodes = TestNode.objects.all()
        self.assertTrue(all(x.data == 'hit' for x in nodes))

def zero_count(node):
    node.counter = 0
    node.save()


def add_onehundred(node):
    node.counter += 100
    node.save()


class TestDeferIteration(TestCase):
    def setUp(self):
        super(TestDeferIteration, self).setUp()

        for x in xrange(10):
            TestNode.objects.create(data="TestNode {}".format(x), counter=x+1)

    def test_that_sharding_works(self):
        _shard(
            TestNode,
            TestNode.objects.all().query,
            zero_count,
            shard_size=1,
            queue="default"
        )

        self.assertNumTasksEquals(10)
        self.process_task_queues()
        self.assertEqual(10, TestNode.objects.filter(counter=0).count())

    def test_that_processing_works(self):
        " By adding 100, we can check that everything is called exactly once"

        defer_iteration(TestNode.objects.all(), add_onehundred)

        self.assertTrue(sum(TestNode.objects.values_list("counter", flat=True)))

        self.process_task_queues()

        self.assertEqual(
            sum([(x+1) + 100 for x in xrange(10)]),
            sum(TestNode.objects.values_list("counter", flat=True))
        )




AUTHENTICATION_BACKENDS = (
    'djangae.contrib.gauth.datastore.backends.AppEngineUserAPIBackend',
)

AUTH_USER_MODEL = 'djangae.GaeDatastoreUser'
LOGIN_URL = 'djangae_login_redirect'

# Set this to True to allow unknown Google users to sign in. Matching is done
# by email. Defaults to False.
# DJANGAE_CREATE_UNKNOWN_USER = False



from django.conf.urls import url

from djangae.contrib.gauth import views


urlpatterns = [
    url(r'^login_redirect$', views.login_redirect, name='djangae_login_redirect'),
    url(r'^switch_accounts/$', views.switch_accounts, name='djangae_switch_accounts'),
]



import datetime
import logging
import contextlib

from django.core.exceptions import NON_FIELD_ERRORS

from google.appengine.ext import db
from google.appengine.api.datastore import Key, Delete
from google.appengine.datastore.datastore_rpc import TransactionOptions

from .unique_utils import unique_identifiers_from_entity
from .utils import key_exists
from djangae.db.backends.appengine.dbapi import IntegrityError, NotSupportedError
from django.conf import settings

DJANGAE_LOG = logging.getLogger("djangae")


def has_active_unique_constraints(model_or_instance):
    """
        Returns true if the model/instance has unique fields or unique_together fields and unique
        constraint checking is enabled on the model
    """

    django_opts = getattr(model_or_instance, "_meta", None)

    # If there are no unique fields on the model, return false
    if not django_opts.unique_together and not any(x.unique for x in django_opts.fields):
        return False

    opts = getattr(model_or_instance, "Djangae", None)
    if opts:
        if hasattr(opts, "disable_constraint_checks"):
            if opts.disable_constraint_checks:
                return False
            else:
                return True

    return not getattr(settings, "DJANGAE_DISABLE_CONSTRAINT_CHECKS", False)



class KeyProperty(db.Property):
    """A property that stores a datastore.Key reference to another object.
        Think of this as a Django GenericForeignKey which returns only the PK value, not the whole
        object, or a db.ReferenceProperty which can point to any model kind, and only returns the Key.
    """

    def validate(self, value):
        if value is None or isinstance(value, Key):
            return value
        raise ValueError("KeyProperty only accepts datastore.Key or None")


class UniqueMarker(db.Model):
    instance = KeyProperty()
    created = db.DateTimeProperty(required=True, auto_now_add=True)

    @staticmethod
    def kind():
        return "_djangae_unique_marker"


@db.transactional(propagation=TransactionOptions.INDEPENDENT, xg=True)
def acquire_identifiers(identifiers, entity_key):
    return _acquire_identifiers(identifiers, entity_key)


def _acquire_identifiers(identifiers, entity_key):
    # This must always be in a cross-group transaction, because even if there's only 1 identifider,
    # in the case where that identifier already exists, we then check if its `instance` exists
    assert entity_key
    namespace = entity_key.namespace() or None
    identifier_keys = [
        Key.from_path(UniqueMarker.kind(), identifier, namespace=namespace) for identifier in identifiers
    ]
    existing_markers = UniqueMarker.get(identifier_keys)
    markers_to_create = []
    markers = []

    for identifier_key, existing_marker in zip(identifier_keys, existing_markers):

        # Backwards compatability: we used to create the markers first in an independent transaction
        # and then create the entity and update the `instance` on the markers.  This meant that it
        # was possible that the independent marker creation transaction finished first and the outer
        # transaction failed, causing stale markers to be left behind.  We no longer do it this way
        # but we still want to ignore any old stale markers, hence if instance is None we overwrite.
        now = datetime.datetime.utcnow()
        if not existing_marker or existing_marker.instance is None:
            markers_to_create.append(UniqueMarker(
                key=identifier_key,
                instance=entity_key,
                created=now
            ))
        elif existing_marker.instance != entity_key and key_exists(existing_marker.instance):
                fields_and_values = identifier.split("|")
                table_name = fields_and_values[0]
                fields_and_values = fields_and_values[1:]
                fields = [ x.split(":")[0] for x in fields_and_values ]
                raise IntegrityError("Unique constraint violation for kind {} on fields: {}".format(table_name, ", ".join(fields)))
        elif existing_marker.instance != entity_key:
            markers_to_create.append(UniqueMarker(
                key=identifier_key,
                instance=entity_key,
                created=now
            ))
        else:
            # The marker is ours anyway
            markers.append(existing_marker)

    db.put(markers_to_create)
    return markers + markers_to_create


def get_markers_for_update(model, old_entity, new_entity):
    """
        Given an old entity state, and the new state, updates the identifiers
        appropriately. Should be called before saving the new_state
    """
    old_ids = set(unique_identifiers_from_entity(model, old_entity, ignore_pk=True))
    new_ids = set(unique_identifiers_from_entity(model, new_entity, ignore_pk=True))

    to_release = old_ids - new_ids
    to_acquire = new_ids - old_ids

    return to_acquire, to_release


def update_instance_on_markers(entity, markers):
    # TODO: fix me!
    def update(marker, instance):
        marker = UniqueMarker.get(marker.key())
        if not marker:
            return

        marker.instance = instance
        marker.put()

    @db.transactional(propagation=TransactionOptions.INDEPENDENT, xg=True)
    def update_all():
        instance = entity.key()
        for marker in markers:
            update(marker, instance)
    update_all()


def acquire(model, entity):
    """
        Given a model and entity, this tries to acquire unique marker locks for the instance. If
        the locks already exist then an IntegrityError will be thrown.
    """
    identifiers = unique_identifiers_from_entity(model, entity, ignore_pk=True)
    return acquire_identifiers(identifiers, entity.key())


def release_markers(markers):
    """ Delete the given UniqueMarker objects. """
    # Note that these should all be from the same Django model instance, and therefore there should
    # be a maximum of 25 of them (because everything blows up if you have more than that - limitation)

    @db.transactional(propagation=TransactionOptions.INDEPENDENT, xg=len(markers) > 1)
    def txn():
        Delete([marker.key() for marker in markers])
    txn()


def release_identifiers(identifiers, namespace):

    @db.transactional(propagation=TransactionOptions.INDEPENDENT, xg=len(identifiers) > 1)
    def txn():
        _release_identifiers(identifiers, namespace)
    txn()


def _release_identifiers(identifiers, namespace):
    keys = [Key.from_path(UniqueMarker.kind(), x, namespace=namespace) for x in identifiers]
    Delete(keys)


def release(model, entity):
    """ Delete the UniqueMarker objects for the given entity. """
    if not has_active_unique_constraints(model):
        return
    identifiers = unique_identifiers_from_entity(model, entity, ignore_pk=True)
    # Key.from_path expects None for an empty namespace, but Key.namespace() returns ''
    namespace = entity.key().namespace() or None
    release_identifiers(identifiers, namespace=namespace)


@db.transactional(propagation=TransactionOptions.INDEPENDENT, xg=True)
def update_identifiers(to_acquire, to_release, key):
    """ A combination of acquire_identifiers and release_identifiers in a combined transaction. """
    _acquire_identifiers(to_acquire, key)
    _release_identifiers(to_release, key.namespace() or None)


class UniquenessMixin(object):
    """ Mixin overriding the methods checking value uniqueness.

    For models defining unique constraints this mixin should be inherited from.
    When iterable (list or set) fields are marked as unique it must be used.
    This is a copy of Django's implementation, save for the part marked by the comment.
    """
    def _perform_unique_checks(self, unique_checks):
        errors = {}
        for model_class, unique_check in unique_checks:
            lookup_kwargs = {}
            for field_name in unique_check:
                f = self._meta.get_field(field_name)
                lookup_value = getattr(self, f.attname)
                if lookup_value is None:
                    continue
                if f.primary_key and not self._state.adding:
                    continue

                ##########################################################################
                # This is a modification to Django's native implementation of this method;
                # we conditionally build a __in lookup if the value is an iterable.
                lookup = str(field_name)
                if isinstance(lookup_value, (list, set, tuple)):
                    lookup = "%s__in" % lookup

                lookup_kwargs[lookup] = lookup_value
                ##########################################################################
                # / end of changes

            if len(unique_check) != len(lookup_kwargs):
                continue

            #######################################################
            # Deal with long __in lookups by doing multiple queries in that case
            # This is a bit hacky, but we really have no choice due to App Engine's
            # 30 multi-query limit. This also means we can't support multiple list fields in
            # a unique combination
            #######################################################

            if len([x for x in lookup_kwargs if x.endswith("__in") ]) > 1:
                raise NotSupportedError("You cannot currently have two list fields in a unique combination")

            # Split IN queries into multiple lookups if they are too long
            lookups = []
            for k, v in lookup_kwargs.iteritems():
                if k.endswith("__in") and len(v) > 30:
                    v = list(v)
                    while v:
                        new_lookup = lookup_kwargs.copy()
                        new_lookup[k] = v[:30]
                        v = v[30:]
                        lookups.append(new_lookup)
                    break
            else:
                # Otherwise just use the one lookup
                lookups = [ lookup_kwargs ]

            for lookup_kwargs in lookups:
                qs = model_class._default_manager.filter(**lookup_kwargs).values_list("pk", flat=True)
                model_class_pk = self._get_pk_val(model_class._meta)
                result = list(qs)

                if not self._state.adding and model_class_pk is not None:
                    # If we are saving an instance, we ignore it's PK in the result
                    try:
                        result.remove(model_class_pk)
                    except ValueError:
                        pass

                if result:
                    if len(unique_check) == 1:
                        key = unique_check[0]
                    else:
                        key = NON_FIELD_ERRORS
                    errors.setdefault(key, []).append(self.unique_error_message(model_class, unique_check))
                    break
        return errors



import warnings

from django.db.utils import IntegrityError
from django.db import transaction
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import BaseUserManager
from django.core.exceptions import ImproperlyConfigured
from django.contrib.auth.backends import ModelBackend
from django.utils import timezone

from google.appengine.api import users

# DJANGAE
from djangae.contrib.gauth.common.models import GaeAbstractBaseUser

# Backwards compatibility, remove before 1.0
# This is here so that we only log once on import, not on each authentication
if hasattr(settings, "ALLOW_USER_PRE_CREATION"):
    warnings.warn(
        "settings.ALLOW_USER_PRE_CREATION is deprecated, "
        "please use DJANGAE_CREATE_UNKNOWN_USER instead"
    )

if hasattr(settings, 'DJANGAE_FORCE_USER_PRE_CREATION'):
    warnings.warn(
        'settings.DJANGAE_FORCE_USER_PRE_CREATION is deprecated, please use'
        ' DJANGAE_CREATE_UNKNOWN_USER instead'
    )

if hasattr(settings, 'DJANGAE_ALLOW_USER_PRE_CREATION'):
    warnings.warn(
        'settings.DJANGAE_ALLOW_USER_PRE_CREATION is deprecated, please use'
        ' DJANGAE_CREATE_UNKNOWN_USER instead'
    )


def should_create_unknown_user():
    """Returns True if we should create a Django user for unknown users.

    Default is False.
    """
    if hasattr(settings, 'DJANGAE_CREATE_UNKNOWN_USER'):
        return settings.DJANGAE_CREATE_UNKNOWN_USER

    if hasattr(settings, 'DJANGAE_FORCE_USER_PRE_CREATION'):
        # This setting meant that there _had_ to be an existing user, it would
        # refuse to create a new user (except for admins).
        return not settings.DJANGAE_FORCE_USER_PRE_CREATION

    if hasattr(settings, 'DJANGAE_ALLOW_USER_PRE_CREATION'):
        return settings.DJANGAE_ALLOW_USER_PRE_CREATION

    if hasattr(settings, 'ALLOW_USER_PRE_CREATION'):
        return settings.ALLOW_USER_PRE_CREATION

    return False


class BaseAppEngineUserAPIBackend(ModelBackend):
    atomic = transaction.atomic
    atomic_kwargs = {}

    def authenticate(self, google_user=None):
        """
        Handles authentication of a user from the given credentials.
        Credentials must be a 'google_user' as returned by the App Engine
        Users API.
        """
        if google_user is None:
            return None

        User = get_user_model()

        if not issubclass(User, GaeAbstractBaseUser):
            raise ImproperlyConfigured(
                "AppEngineUserAPIBackend requires AUTH_USER_MODEL to be a "
                " subclass of djangae.contrib.auth.base.GaeAbstractBaseUser."
            )

        user_id = google_user.user_id()
        email = BaseUserManager.normalize_email(google_user.email())

        try:
            # User exists and we can bail immediately.
            return User.objects.get(username=user_id)
        except User.DoesNotExist:
            pass

        auto_create = should_create_unknown_user()
        user_is_admin = users.is_current_user_admin()

        if not (auto_create or user_is_admin):
            # User doesn't exist and we aren't going to create one.
            return None

        # OK. We will grant access. We may need to update an existing user, or
        # create a new one, or both.

        # Those 3 scenarios are:
        # 1. A User object has been created for this user, but that they have not logged in yet.
        # In this case wefetch the User object by email, and then update it with the Google User ID
        # 2. A User object exists for this email address but belonging to a different Google account.
        # This generally only happens when the email address of a Google Apps account has been
        # signed up as a Google account and then the apps account itself has actually become a
        # Google account. This is possible but very unlikely.
        # 3. There is no User object realting to this user whatsoever.

        try:
            existing_user = User.objects.get(email=email)
        except User.DoesNotExist:
            existing_user = None

        if existing_user:
            if existing_user.username is None:
                # We can use the existing user for this new login.
                existing_user.username = user_id
                existing_user.email = email
                existing_user.last_login = timezone.now()
                existing_user.save()

                return existing_user
            else:
                # We need to update the existing user and create a new one.
                with self.atomic(**self.atomic_kwargs):
                    existing_user = User.objects.get(pk=existing_user.pk)
                    existing_user.email = None
                    existing_user.save()

                    return User.objects.create_user(user_id, email=email)
        else:
            # Create a new user, but account for another thread having created it already in a race
            # condition scenario. Our logic cannot be in a transaction, so we have to just catch this.
            try:
                return User.objects.create_user(user_id, email=email)
            except IntegrityError:
                return User.objects.get(username=user_id)



import django
from django.db import DEFAULT_DB_ALIAS, router
from django.apps import apps
from django.utils.encoding import smart_text
from django.utils import six
from django.utils.six.moves import input


def update_contenttypes(sender, verbosity=2, db=DEFAULT_DB_ALIAS, **kwargs):
    """
        Django's default update_contenttypes relies on many inconsistent queries which causes problems
        with syncdb. This monkeypatch replaces it with a version that does look ups on unique constraints
        which are slightly better protected from eventual consistency issues by the context cache.
    """
    from django.contrib.contenttypes.models import ContentType
    
    if verbosity >= 2:
        print("Running Djangae version of update_contenttypes on {}".format(sender))

    try:
        apps.get_model('contenttypes', 'ContentType')
    except LookupError:
        return

    if hasattr(router, "allow_migrate_model"):
        if not router.allow_migrate_model(db, ContentType):
            return

    ContentType.objects.clear_cache()
    app_models = sender.get_models()
    if not app_models:
        return
    # They all have the same app_label, get the first one.
    app_label = sender.label
    app_models = dict(
        (model._meta.model_name, model)
        for model in app_models
    )

    created_or_existing_pks = []
    created_or_existing_by_unique = {}

    for (model_name, model) in six.iteritems(app_models):
        # Go through get_or_create any models that we want to keep
        defaults = {}
        if django.VERSION < (1, 9):
            defaults['name'] = smart_text(model._meta.verbose_name_raw)

        ct, created = ContentType.objects.get_or_create(
            app_label=app_label,
            model=model_name,
            defaults=defaults,
        )

        if verbosity >= 2 and created:
            print("Adding content type '%s | %s'" % (ct.app_label, ct.model))

        created_or_existing_pks.append(ct.pk)
        created_or_existing_by_unique[(app_label, model_name)] = ct.pk

    # Now lets see if we should remove any

    to_remove = [x for x in ContentType.objects.filter(app_label=app_label) if x.pk not in created_or_existing_pks]

    # Now it's possible that our get_or_create failed because of consistency issues and we create a duplicate.
    # Then the original appears in the to_remove and we remove the original. This is bad. So here we go through the
    # to_remove list, and if we created the content type just now, we delete that one, and restore the original in the
    # cache
    for ct in to_remove:
        unique = (ct.app_label, ct.model)
        if unique in created_or_existing_by_unique:
            # We accidentally created a duplicate above due to HRD issues, delete the one we created
            ContentType.objects.get(pk=created_or_existing_by_unique[unique]).delete()
            created_or_existing_by_unique[unique] = ct.pk
            ct.save()  # Recache this one in the context cache

    to_remove = [ x for x in to_remove if (x.app_label, x.model) not in created_or_existing_by_unique ]

    # Now, anything left should actually be a stale thing. It's still possible we missed some but they'll get picked up
    # next time. Confirm that the content type is stale before deletion.
    if to_remove:
        if kwargs.get('interactive', False):
            content_type_display = '\n'.join([
                '    %s | %s' % (x.app_label, x.model)
                for x in to_remove
            ])
            ok_to_delete = input("""The following content types are stale and need to be deleted:

%s

Any objects related to these content types by a foreign key will also
be deleted. Are you sure you want to delete these content types?
If you're unsure, answer 'no'.

    Type 'yes' to continue, or 'no' to cancel: """ % content_type_display)
        else:
            ok_to_delete = False

        if ok_to_delete == 'yes':
            for ct in to_remove:
                if verbosity >= 2:
                    print("Deleting stale content type '%s | %s'" % (ct.app_label, ct.model))
                ct.delete()
        else:
            if verbosity >= 2:
                print("Stale content types remain.")



import django
import json
import logging
import re
import datetime

from itertools import chain, imap
from django.core.exceptions import FieldError
from django.db.models.fields import FieldDoesNotExist
from django.db.models.sql.datastructures import EmptyResultSet

from django.db.models import AutoField

from django.db.models.query import QuerySet

try:
    from django.db.models.query import FlatValuesListIterable
except ImportError:
    # Django < 1.8
    class FlatValuesListIterable(object):
        pass

try:
    from django.db.models.query import ValuesListQuerySet
except ImportError:
    # Django >= 1.9
    class ValuesListQuerySet(object):
        pass

from django.db.models.expressions import Star

from django.db import NotSupportedError
from djangae.db.backends.appengine.indexing import (
    REQUIRES_SPECIAL_INDEXES,
    add_special_index,
)

from djangae import environment
from djangae.db.backends.appengine import POLYMODEL_CLASS_ATTRIBUTE
from djangae.db.utils import (
    get_top_concrete_parent,
    has_concrete_parents,
    get_field_from_column
)

from google.appengine.api import datastore


DJANGAE_LOG = logging.getLogger("djangae")


VALID_QUERY_KINDS = (
    "SELECT",
    "UPDATE",
    "INSERT",
    "DELETE",
    "COUNT",
    "AVERAGE"
)

VALID_ANNOTATIONS = {
    "MIN": min,
    "MAX": max,
    "SUM": sum,
    "COUNT": len,
    "AVG": lambda x: (sum(x) / len(x))
}

VALID_CONNECTORS = (
    'AND', 'OR'
)


VALID_OPERATORS = (
    '=', '<', '>', '<=', '>=', 'IN'
)

def convert_operator(operator):
    if operator == 'exact':
        return '='
    elif operator == 'gt':
        return '>'
    elif operator == 'lt':
        return '<'
    elif operator == 'gte':
        return '>='
    elif operator == 'lte':
        return '<='

    return operator.upper()

class WhereNode(object):
    def __init__(self):
        self.column = None
        self.operator = None
        self.value = None
        self.output_field = None

        self.children = []
        self.connector = 'AND'
        self.negated = False

    @property
    def is_leaf(self):
        return bool(self.column and self.operator)

    def set_connector(self, connector):
        self.connector = connector

    def append_child(self, node):
        self.children.append(node)

    def set_leaf(self, column, operator, value, is_pk_field, negated, namespace, target_field=None):
        assert column
        assert operator
        assert isinstance(is_pk_field, bool)
        assert isinstance(negated, bool)

        if operator == "iexact" and isinstance(target_field, AutoField):
            # When new instance is created, automatic primary key 'id' does not generate '_idx_iexact_id'.
            # As the primary key 'id' (AutoField) is integer and is always case insensitive,
            # we can deal with 'id_iexact=' query by using 'exact' rather than 'iexact'.
            operator = "exact"
            value = int(value)

        if is_pk_field:
            # If this is a primary key, we need to make sure that the value
            # we pass to the query is a datastore Key. We have to deal with IN queries here
            # because they aren't flattened until the DNF stage
            model = get_top_concrete_parent(target_field.model)
            table = model._meta.db_table

            if isinstance(value, (list, tuple)):
                value = [
                    datastore.Key.from_path(table, x, namespace=namespace)
                    for x in value if x
                ]
            else:
                if operator == "isnull" and value is True:
                    # FIXME: Strictly, this isn't correct, this could be one of several branches
                    # but id=None filters are silly anyway. This should be moved to after normalization..
                    # probably. This fixes a test in Django which does this in get_or_create for some reason
                    raise EmptyResultSet()

                if not value:
                    # Empty strings and 0 are forbidden as keys
                    # so make this an impossible filter
                    # FIXME: This is a hack! It screws with the ordering
                    # because it's an inequality. Instead we should wipe this
                    # filter out when preprocessing in the DNF (because it's impossible)
                    value = datastore.Key.from_path('', 1)
                    operator = '<'
                else:
                    value = datastore.Key.from_path(table, value, namespace=namespace)
            column = "__key__"

        # Do any special index conversions necessary to perform this lookup
        primary_operation = operator.split("__")[0]
        special_indexer = REQUIRES_SPECIAL_INDEXES.get(primary_operation)

        if special_indexer:
            if is_pk_field:
                column = model._meta.pk.column
                value = unicode(value.id_or_name())

            add_special_index(target_field.model, column, primary_operation, value)
            index_type = special_indexer.prepare_index_type(operator, value)
            value = special_indexer.prep_value_for_query(value)
            column = special_indexer.indexed_column_name(column, value, index_type)
            operator = special_indexer.prep_query_operator(operator)

        self.column = column
        self.operator = convert_operator(operator)
        self.value = value



    def __iter__(self):
        for child in chain(*imap(iter, self.children)):
            yield child
        yield self


    def __repr__(self):
        if self.is_leaf:
            return "[%s%s%s]" % (self.column, self.operator, self.value)
        else:
            return "(%s:%s%s)" % (self.connector, "!" if self.negated else "", ",".join([repr(x) for x in self.children]))

    def __eq__(self, rhs):
        if self.is_leaf != rhs.is_leaf:
            return False

        if self.is_leaf:
            return self.column == rhs.column and self.value == rhs.value and self.operator == rhs.operator
        else:
            return self.connector == rhs.connector and self.children == rhs.children

    def __hash__(self):
        if self.is_leaf:
            return hash((self.column, self.value, self.operator))
        else:
            return hash((self.connector,) + tuple([hash(x) for x in self.children]))

class Query(object):
    def __init__(self, model, kind):
        assert kind in VALID_QUERY_KINDS

        self.model = model
        self.concrete_model = get_top_concrete_parent(model)
        self.kind = kind

        self.projection_possible = True
        self.tables = []

        self.columns = None # None means all fields
        self.init_list = []

        self.distinct = False
        self.order_by = []
        self.row_data = [] # For insert/updates
        self._where = None
        self.low_mark = self.high_mark = None

        self.annotations = []
        self.per_entity_annotations = []
        self.extra_selects = []
        self.polymodel_filter_added = False

        # A list of PKs that should be excluded from the resultset
        self.excluded_pks = set()

    @property
    def is_normalized(self):
        """
            Returns True if this query has a normalized where tree
        """
        if not self.where:
            return True

        # Only a leaf node, return True
        if not self.where.is_leaf:
            return True

        # If we have children, and they are all leaf nodes then this is a normalized
        # query
        return self.where.connector == 'OR' and self.where.children and all(x.is_leaf for x in self.where.children)

    def add_extra_select(self, column, lookup):
        if lookup.lower().startswith("select "):
            raise ValueError("SQL statements aren't supported with extra(select=)")

        # Boolean expression test
        bool_expr = "(?P<lhs>[a-zA-Z0-9_]+)\s?(?P<op>[=|>|<]{1,2})\s?(?P<rhs>[\w+|\']+)"

        # Operator expression test
        op_expr = "(?P<lhs>[a-zA-Z0-9_]+)\s?(?P<op>[+|-|/|*])\s?(?P<rhs>[\w+|\']+)"

        OP_LOOKUP = {
            "=": lambda x, y: x == y,
            "is": lambda x, y: x == y,
            "<": lambda x, y: x < y,
            ">": lambda x, y: x > y,
            ">=": lambda x, y: x >= y,
            "<=": lambda x, y: x <= y,
            "+": lambda x, y: x + y,
            "-": lambda x, y: x - y,
            "/": lambda x, y: x / y,
            "*": lambda x, y: x * y
        }

        for regex in (bool_expr, op_expr):
            match = re.match(regex, lookup)
            if match:
                lhs = match.group('lhs')
                rhs = match.group('rhs')
                op = match.group('op').lower()
                if op in OP_LOOKUP:
                    self.extra_selects.append((column, (OP_LOOKUP[op], (lhs, rhs))))
                else:
                    raise ValueError("Unsupported operator")
                return

        # Assume literal
        self.extra_selects.append((column, (lambda x: x, [lookup])))

    def add_source_table(self, table):
        if table in self.tables:
            return

        self.tables.append(table)

    def set_distinct(self, distinct_fields):
        self.distinct = True
        if distinct_fields:
            for field in distinct_fields:
                self.add_projected_column(field)
        elif not self.columns:
            for field in self.model._meta.fields:
                self.add_projected_column(field.column)

    def add_order_by(self, column):
        self.order_by.append(column)

    def add_annotation(self, column, annotation):
        name = annotation.__class__.__name__
        if name == "Count":
            return # Handled elsewhere

        if name not in ("Col", "Date", "DateTime"):
            raise NotSupportedError("Unsupported annotation %s" % name)

        def process_date(value, lookup_type):
            from djangae.db.backends.appengine.commands import ensure_datetime #FIXME move func to utils
            value = ensure_datetime(value)
            ret = datetime.datetime.fromtimestamp(0)

            POSSIBLE_LOOKUPS = ("year", "month", "day", "hour", "minute", "second")

            ret = ret.replace(
                value.year,
                value.month if lookup_type in POSSIBLE_LOOKUPS[1:] else ret.month,
                value.day if lookup_type in POSSIBLE_LOOKUPS[2:] else ret.day,
                value.hour if lookup_type in POSSIBLE_LOOKUPS[3:] else ret.hour,
                value.minute if lookup_type in POSSIBLE_LOOKUPS[4:] else ret.minute,
                value.second if lookup_type in POSSIBLE_LOOKUPS[5:] else ret.second,
            )

            return ret

        # Abuse the extra_select functionality
        if name == "Col":
            self.extra_selects.append((column, (lambda x: x, [column])))
        elif name in ("Date", "DateTime"):
            self.extra_selects.append(
                (column,
                (lambda x: process_date(x, annotation.lookup_type), [getattr(annotation, "lookup", column)]))
            )
            # Override the projection so that we only get this column
            self.columns = set([ getattr(annotation, "lookup", column) ])


    def add_projected_column(self, column):
        self.init_list.append(column)

        if not self.projection_possible:
            # If we previously tried to add a column that couldn't be
            # projected, we don't try and add any more
            return

        field = get_field_from_column(self.model, column)

        if field is None:
            raise NotSupportedError("{} is not a valid column for the queried model. Did you try to join?".format(column))

        if field.db_type(self.connection) in ("bytes", "text", "list", "set"):
            DJANGAE_LOG.warn("Disabling projection query as %s is an unprojectable type", column)
            self.columns = None
            self.projection_possible = False
            return

        if not self.columns:
            self.columns = set([ column ])
        else:
            self.columns.add(column)

    def add_row(self, data):
        assert self.columns
        assert len(data) == len(self.columns)

        self.row_data.append(data)

    def prepare(self):
        if not self.init_list:
            self.init_list = [ x.column for x in self.model._meta.fields ]

        self._remove_erroneous_isnull()
        self._remove_negated_empty_in()
        self._add_inheritence_filter()
        self._populate_excluded_pks()
        self._disable_projection_if_fields_used_in_equality_filter()
        self._check_only_single_inequality_filter()

    @property
    def where(self):
        return self._where

    @where.setter
    def where(self, where):
        assert where is None or isinstance(where, WhereNode)

        self._where = where

    def _populate_excluded_pks(self):
        if not self._where:
            return

        self.excluded_pks = set()
        def walk(node, negated):
            if node.connector == "OR":
                # We can only process AND nodes, if we hit an OR we can't
                # use the excluded PK optimization
                return

            if node.negated:
                negated = not negated

            for child in node.children[:]:
                if negated and child.operator == "=" and child.column == "__key__":
                    self.excluded_pks.add(child.value)
                    node.children.remove(child)
                elif negated and child.operator == "IN" and child.column == "__key__":
                    [ self.excluded_pks.add(x) for x in child.value ]
                    node.children.remove(child)

                walk(child, negated)

            node.children = [ x for x in node.children if x.children or x.column ]

        walk(self._where, False)

        if not self._where.children:
            self._where = None

    def _remove_negated_empty_in(self):
        """
            An empty exclude(id__in=[]) is pointless, but can cause trouble
            during denormalization. We remove such nodes here.
        """
        if not self._where:
            return

        def walk(node, negated):
            if node.negated:
                negated = node.negated

            for child in node.children[:]:
                if negated and child.operator == "IN" and not child.value:
                    node.children.remove(child)

                walk(child, negated)

            node.children = [ x for x in node.children if x.children or x.column ]

        had_where = bool(self._where.children)
        walk(self._where, False)

        # Reset the where if that was the only filter
        if had_where and not bool(self._where.children):
            self._where = None

    def _remove_erroneous_isnull(self):
        # This is a little crazy, but bear with me...
        # If you run a query like this:  filter(thing=1).exclude(field1="test") where field1 is
        # null-able you'll end up with a negated branch in the where tree which is:

        #           AND (negated)
        #          /            \
        #   field1="test"   field1__isnull=False

        # This is because on SQL, field1 != "test" won't give back rows where field1 is null, so
        # django has to include the negated isnull=False as well in order to get back the null rows
        # as well.  On App Engine though None is just a value, not the lack of a value, so it's
        # enough to just have the first branch in the negated node and in fact, if you try to use
        # the above tree, it will result in looking for:
        #  field1 < "test" and field1 > "test" and field1__isnull=True
        # which returns the wrong result (it'll return when field1 == None only)

        def walk(node, negated):
            if node.negated:
                negated = not negated

            if not node.is_leaf:
                equality_fields = set()
                negated_isnull_fields = set()
                isnull_lookup = {}

                for child in node.children[:]:
                    if negated:
                        if child.operator == "=":
                            equality_fields.add(child.column)
                            if child.column in negated_isnull_fields:
                                node.children.remove(isnull_lookup[child.column])

                        elif child.operator == "ISNULL":
                            negated_isnull_fields.add(child.column)
                            if child.column in equality_fields:
                                node.children.remove(child)
                            else:
                                isnull_lookup[child.column] = child

                    walk(child, negated)
        if self.where:
            walk(self._where, False)

    def _check_only_single_inequality_filter(self):
        inequality_fields = set()
        def walk(node, negated):
            if node.negated:
                negated = not negated

            for child in node.children[:]:
                if (negated and child.operator == "=") or child.operator in (">", "<", ">=", "<="):
                    inequality_fields.add(child.column)
                walk(child, negated)

            if len(inequality_fields) > 1:
                raise NotSupportedError(
                    "You can only have one inequality filter per query on the datastore. "
                    "Filters were: %s" % ' '.join(inequality_fields)
                )

        if self.where:
            walk(self._where, False)

    def _disable_projection_if_fields_used_in_equality_filter(self):
        if not self._where or not self.columns:
            return

        equality_columns = set()

        def walk(node):
            if not node.is_leaf:
                for child in node.children:
                    walk(child)
            elif node.operator == "=" or node.operator == "IN":
                equality_columns.add(node.column)

        walk(self._where)

        if equality_columns and equality_columns.intersection(self.columns):
            self.columns = None
            self.projection_possible = False

    def _add_inheritence_filter(self):
        """
            We support inheritence with polymodels. Whenever we set
            the 'where' on this query, we manipulate the tree so that
            the lookups are ANDed with a filter on 'class = db_table'
            and on inserts, we add the 'class' column if the model is part
            of an inheritance tree.

            We only do any of this if the model has concrete parents and isn't
            a proxy model
        """
        if has_concrete_parents(self.model) and not self.model._meta.proxy:
            if self.polymodel_filter_added:
                return

            new_filter = WhereNode()
            new_filter.column = POLYMODEL_CLASS_ATTRIBUTE
            new_filter.operator = '='
            new_filter.value = self.model._meta.db_table

            # We add this bare AND just to stay consistent with what Django does
            new_and = WhereNode()
            new_and.connector = 'AND'
            new_and.children = [ new_filter ]

            new_root = WhereNode()
            new_root.connector = 'AND'
            new_root.children = [ new_and ]
            if self._where:
                # Add the original where if there was one
                new_root.children.append(self._where)
            self._where = new_root

            self.polymodel_filter_added = True

    def serialize(self):
        """
            The idea behind this function is to provide a way to serialize this
            query to a string which can be compared to another query. Pickle won't
            work as some of the values etc. might not be picklable.

            FIXME: This function is incomplete! Not all necessary members are serialized
        """
        if not self.is_normalized:
            raise ValueError("You cannot serialize queries unless they are normalized")

        result = {}
        result["kind"] = self.kind
        result["table"] = self.model._meta.db_table
        result["concrete_table"] = self.concrete_model._meta.db_table
        result["columns"] = list(self.columns or []) # set() is not JSONifiable
        result["projection_possible"] = self.projection_possible
        result["init_list"] = self.init_list
        result["distinct"] = self.distinct
        result["order_by"] = self.order_by
        result["low_mark"] = self.low_mark
        result["high_mark"] = self.high_mark
        result["excluded_pks"] = map(str, self.excluded_pks)

        where = []

        if self.where:
            assert self.where.connector == 'OR'
            for node in self.where.children:
                assert node.connector == 'AND'

                query = {}
                for lookup in node.children:
                    query[''.join([lookup.column, lookup.operator])] = unicode(lookup.value)

                where.append(query)

        result["where"] = where

        return json.dumps(result)


INVALID_ORDERING_FIELD_MESSAGE = (
    "Ordering on TextField or BinaryField is not supported on the datastore. "
    "You might consider using a ComputedCharField which stores the first "
    "_MAX_STRING_LENGTH (from google.appengine.api.datastore_types) bytes of the "
    "field and instead order on that."
)


def _extract_ordering_from_query_18(query):
    from djangae.db.backends.appengine.commands import log_once
    from django.db.models.expressions import OrderBy, F

    # Add any orderings
    if not query.default_ordering:
        result = list(query.order_by)
    else:
        result = list(query.order_by or query.get_meta().ordering or [])

    if query.extra_order_by:
        result = list(query.extra_order_by)

        # we need some extra logic to handle dot seperated ordering
        new_result = []
        cross_table_ordering = set()
        for ordering in result:
            if "." in ordering:
                dot_based_ordering = ordering.split(".")
                if dot_based_ordering[0] == query.model._meta.db_table:
                    ordering = dot_based_ordering[1]
                elif dot_based_ordering[0].lstrip('-') == query.model._meta.db_table:
                    ordering = '-{}'.format(dot_based_ordering[1])
                else:
                    cross_table_ordering.add(ordering)
                    continue # we don't want to add this ordering value
            new_result.append(ordering)

        if len(cross_table_ordering):
            log_once(
                DJANGAE_LOG.warning if environment.is_development_environment() else DJANGAE_LOG.debug,
                "The following orderings were ignored as cross-table orderings are not supported on the datastore: %s", cross_table_ordering
            )

        result = new_result

    final = []

    opts = query.model._meta

    # Apparently expression ordering is absolute and so shouldn't be flipped
    # if the standard_ordering is False. This keeps track of which columns
    # were expressions and so don't need flipping
    expressions = set()

    for col in result:
        if isinstance(col, OrderBy):
            descending = col.descending
            col = col.expression.name
            if descending:
                col = "-" + col
            expressions.add(col)

        elif isinstance(col, F):
            col = col.name

        if isinstance(col, (int, long)):
            # If you do a Dates query, the ordering is set to [1] or [-1]... which is weird
            # I think it's to select the column number but then there is only 1 column so
            # unless the ordinal is one-based I have no idea. So basically if it's an integer
            # subtract 1 from the absolute value and look up in the select for the column (guessing)
            idx = abs(col) - 1
            try:
                field_name = query.select[idx].col.col[-1]
                field = query.model._meta.get_field(field_name)
                final.append("-" + field.column if col < 0 else field.column)
            except IndexError:
                raise NotSupportedError("Unsupported order_by %s" % col)
        elif col.lstrip("-") == "pk":
            pk_col = "__key__"
            final.append("-" + pk_col if col.startswith("-") else pk_col)
        elif col == "?":
            raise NotSupportedError("Random ordering is not supported on the datastore")
        elif col.lstrip("-").startswith("__") and col.endswith("__"):
            # Allow stuff like __scatter__
            final.append(col)
        elif "__" in col:
            continue
        else:
            try:
                column = col.lstrip("-")

                # This is really 1.8 only, but I didn't want to duplicate this function
                # just for this. Suggestions for doing this more cleanly welcome!
                if column in getattr(query, "annotation_select", {}):
                    # It's an annotation, if it's a supported one, return the
                    # original column
                    annotation = query.annotation_select[column]
                    name = annotation.__class__.__name__

                    # We only support a few expressions
                    if name not in ("Col", "Date", "DateTime"):
                        raise NotSupportedError("Tried to order by unsupported expression")
                    else:
                        # Retrieve the original column and use that for ordering
                        if name == "Col":
                            column = annotation.output_field.column
                        else:
                            column = annotation.col.output_field.column

                field = query.model._meta.get_field(column)

                if field.get_internal_type() in (u"TextField", u"BinaryField"):
                    raise NotSupportedError(INVALID_ORDERING_FIELD_MESSAGE)

                # If someone orders by 'fk' rather than 'fk_id' this complains as that should take
                # into account the related model ordering. Note the difference between field.name == column
                # and field.attname (X_id)
                if field.related_model and field.name == column and field.related_model._meta.ordering:
                    raise NotSupportedError("Related ordering is not supported on the datastore")

                column = "__key__" if field.primary_key else field.column
                final.append("-" + column if col.startswith("-") else column)
            except FieldDoesNotExist:
                if col in query.extra_select:
                    # If the column is in the extra select we transform to the original
                    # column
                    try:
                        field = opts.get_field(query.extra_select[col][0])
                        column = "__key__" if field.primary_key else field.column
                        final.append("-" + column if col.startswith("-") else column)
                        continue
                    except FieldDoesNotExist:
                        # Just pass through to the exception below
                        pass

                available = opts.get_all_field_names()
                raise FieldError("Cannot resolve keyword %r into field. "
                    "Choices are: %s" % (col, ", ".join(available))
                )

    # Reverse if not using standard ordering
    def swap(col):
        if col.startswith("-"):
            return col.lstrip("-")
        else:
            return "-{}".format(col)

    if not query.standard_ordering:
        final = [ x if x in expressions else swap(x) for x in final ]

    if len(final) != len(result):
        diff = set(result) - set(final)
        log_once(
            DJANGAE_LOG.warning if environment.is_development_environment() else DJANGAE_LOG.debug,
            "The following orderings were ignored as cross-table and random orderings are not supported on the datastore: %s", diff
        )

    return final


def _extract_projected_columns_from_query_18(query):
    result = []

    if query.select:
        for x in query.select:
            column = x.target.column
            result.append(column)
        return result
    else:
        # If the query uses defer()/only() then we need to process deferred. We have to get all deferred columns
        # for all (concrete) inherited models and then only include columns if they appear in that list
        only_load = query.get_loaded_field_names()
        if only_load:
            for field, model in query.model._meta.get_concrete_fields_with_model():
                model = model or query.model
                try:
                    if field.column in only_load[model]:
                        # Add a field that has been explicitly included
                        result.append(field.column)
                except KeyError:
                    # Model wasn't explicitly listed in the only_load table
                    # Therefore, we need to load all fields from this model
                    result.append(field.column)
            return result
        else:
            return []


def _walk_django_where(query, trunk_callback, leaf_callback, **kwargs):
    """
        Walks through a Django where tree. If a leaf node is encountered
        the leaf_callback is called, otherwise the trunk_callback is called
    """

    def walk_node(node, **kwargs):
        negated = kwargs["negated"]

        if node.negated:
            negated = not negated

        for child in node.children:
            new_kwargs = kwargs.copy()
            new_kwargs["negated"] = negated
            if not getattr(child, "children", []):
                leaf_callback(child, **new_kwargs)
            else:
                new_parent = trunk_callback(child, **new_kwargs)

                if new_parent:
                    new_kwargs["new_parent"] = new_parent

                walk_node(child, **new_kwargs)

    kwargs.setdefault("negated", False)
    walk_node(query.where, **kwargs)

def _django_18_query_walk_leaf(node, negated, new_parent, connection, model):
    new_node = WhereNode()

    if not hasattr(node, "lhs"):
        raise NotSupportedError("Attempted probable subquery, these aren't supported on the datastore")

    # Leaf
    if hasattr(node.lhs, 'target'):
        # from Django 1.9, some node.lhs might not have a target attribute
        # as they might be wrapping date fields
        field = node.lhs.target
        operator = node.lookup_name
    else:
        field = node.lhs.lhs.target
        operator = node.lhs.lookup_name

        # This deals with things like datefield__month__gt=X which means from this point
        # on, operator will have two parts in that particular case and will probably need to
        # be dealt with by a special indexer
        if node.lookup_name != operator:
            operator = "{}__{}".format(operator, node.lookup_name)

    if get_top_concrete_parent(field.model) != get_top_concrete_parent(model):
        raise NotSupportedError("Cross-join where filters are not supported on the datastore")

    # Make sure we don't let people try to filter on a text field, otherwise they just won't
    # get any results!

    if field.db_type(connection) in ("bytes", "text"):
        raise NotSupportedError("You can't filter on text or blob fields on the datastore")

    if operator == "isnull" and field.model._meta.parents.values():
        raise NotSupportedError("isnull lookups on inherited relations aren't supported on the datastore")

    lhs = field.column

    try:
        if hasattr(node.rhs, "get_compiler"):
            # This is a subquery
            raise NotSupportedError("Attempted to run a subquery on the datastore")
        elif isinstance(node.rhs, ValuesListQuerySet):
            # We explicitly handle ValuesListQuerySet because of the
            # common case of pk__in=Something.objects.values_list("pk", flat=True)
            # this WILL execute another query, but that is to be expected on a
            # non-relational datastore.

            node.rhs = [ x for x in node.rhs ] # Evaluate the queryset
            rhs = node.process_rhs(None, connection) # Process the RHS as if it was a list

        elif isinstance(node.rhs, QuerySet):
            # In Django 1.9, ValuesListQuerySet doesn't exist anymore, and instead
            # values_list returns a QuerySet
            if node.rhs._iterable_class == FlatValuesListIterable:
                # if the queryset has FlatValuesListIterable as iterable class
                # then it's a flat list, and we just need to evaluate the
                # queryset converting it into a list
                node.rhs = [ x for x in node.rhs ]
            else:
                # otherwise, we try to get the PK from the queryset
                node.rhs = [ x.pk for x in node.rhs ]

            rhs = node.process_rhs(None, connection) # Process the RHS as if it was a list

        else:
            rhs = node.process_rhs(None, connection)
    except EmptyResultSet:
        if operator == 'in':
            # Deal with this later
            rhs = [ [] ]
        else:
            raise


    if operator in ('in', 'range'):
        rhs = rhs[-1]
    elif operator == 'isnull':
        rhs = node.rhs
    else:
        rhs = rhs[-1][0]

    new_node.set_leaf(
        lhs,
        operator,
        rhs,
        is_pk_field=field==model._meta.pk,
        negated=negated,
        namespace=connection.ops.connection.settings_dict.get("NAMESPACE"),
        target_field=field,
    )

    # For some reason, this test:
    # test_update_with_related_manager (get_or_create.tests.UpdateOrCreateTests)
    # ends up with duplicate nodes in the where tree. I don't know why. But this
    # weirdly causes the datastore query to return nothing.
    # so here we don't add duplicate nodes, I can't think of a case where that would
    # change the query if it's under the same parent.
    if new_node in new_parent.children:
        return

    new_parent.children.append(new_node)

def _django_18_query_walk_trunk(node, negated, new_parent, **kwargs):
    new_node = WhereNode()
    new_node.connector = node.connector
    new_node.negated = node.negated

    new_parent.children.append(new_node)

    return new_node


def _transform_query_18(connection, kind, query):
    from django.db.models.sql.where import EmptyWhere
    if isinstance(query.where, EmptyWhere):
        # Empty where means return nothing!
        raise EmptyResultSet()

    ret = Query(query.model, kind)
    ret.connection = connection

    # Add the root concrete table as the source table
    root_table = get_top_concrete_parent(query.model)._meta.db_table
    ret.add_source_table(root_table)

    # Extract the ordering of the query results
    for order_col in _extract_ordering_from_query_18(query):
        ret.add_order_by(order_col)

    # Extract any projected columns (values/values_list/only/defer)
    for projected_col in _extract_projected_columns_from_query_18(query):
        ret.add_projected_column(projected_col)

    # Add any extra selects
    for col, select in query.extra_select.items():
        ret.add_extra_select(col, select[0])

    if query.distinct:
        # This must happen after extracting projected cols
        ret.set_distinct(list(query.distinct_fields))

    # Process annotations!
    if query.annotation_select:
        for k, v in query.annotation_select.items():
            ret.add_annotation(k, v)

    # Extract any query offsets and limits
    ret.low_mark = query.low_mark
    ret.high_mark = query.high_mark

    output = WhereNode()
    output.connector = query.where.connector

    _walk_django_where(
        query,
        _django_18_query_walk_trunk,
        _django_18_query_walk_leaf,
        new_parent=output,
        connection=connection,
        negated=query.where.negated,
        model=query.model
    )

    # If there no child nodes, just wipe out the where
    if not output.children:
        output = None

    ret.where = output

    return ret


def _transform_query_19(connection, kind, query):
    from django.db.models.sql.where import NothingNode, WhereNode as DjangoWhereNode

    # It could either be a NothingNode, or a WhereNode(AND NothingNode)
    if (
            isinstance(query.where, NothingNode) or (
            isinstance(query.where, DjangoWhereNode) and
            len(query.where.children) == 1 and
            isinstance(query.where.children[0], NothingNode)
            )):
        # Empty where means return nothing!
        raise EmptyResultSet()

    ret = Query(query.model, kind)
    ret.connection = connection

    # Add the root concrete table as the source table
    root_table = get_top_concrete_parent(query.model)._meta.db_table
    ret.add_source_table(root_table)

    # Extract the ordering of the query results
    for order_col in _extract_ordering_from_query_18(query):
        ret.add_order_by(order_col)

    # Extract any projected columns (values/values_list/only/defer)
    for projected_col in _extract_projected_columns_from_query_18(query):
        ret.add_projected_column(projected_col)

    # Add any extra selects
    for col, select in query.extra_select.items():
        ret.add_extra_select(col, select[0])

    if query.distinct:
        # This must happen after extracting projected cols
        ret.set_distinct(list(query.distinct_fields))

    # Process annotations!
    if query.annotation_select:
        for k, v in query.annotation_select.items():
            ret.add_annotation(k, v)

    # Extract any query offsets and limits
    ret.low_mark = query.low_mark
    ret.high_mark = query.high_mark

    output = WhereNode()
    output.connector = query.where.connector

    _walk_django_where(
        query,
        _django_18_query_walk_trunk,
        _django_18_query_walk_leaf,
        new_parent=output,
        connection=connection,
        negated=query.where.negated,
        model=query.model
    )

    # If there no child nodes, just wipe out the where
    if not output.children:
        output = None

    ret.where = output

    return ret


def get_factory_for_version(factory, version):
    try:
        return factory[version]
    except KeyError:
        return factory[max(factory.keys())]


_FACTORY = {
    (1, 8): _transform_query_18,
    (1, 9): _transform_query_19,
    (1, 10): _transform_query_19, # Same as 1.9
}


def _determine_query_kind_18(query):
    if query.annotations:
        if "__count" in query.annotations:
            field = query.annotations["__count"].input_field
            if isinstance(field, Star) or field.value == "*":
                return "COUNT"

    return "SELECT"


def _determine_query_kind_19(query):
    if query.annotations:
        if "__count" in query.annotations:
            field = query.annotations["__count"].source_expressions[0]
            if isinstance(field, Star) or field.value == "*":
                return "COUNT"

    return "SELECT"


_KIND_FACTORY = {
    (1, 8): _determine_query_kind_18,
    (1, 9): _determine_query_kind_19,
    (1, 10): _determine_query_kind_19  # Same as 1.9
}


def transform_query(connection, query):
    version = django.VERSION[:2]
    kind = get_factory_for_version(_KIND_FACTORY, version)(query)
    return get_factory_for_version(_FACTORY, version)(connection, kind, query)


_ORDERING_FACTORY = {
    (1, 8): _extract_ordering_from_query_18,
    (1, 9): _extract_ordering_from_query_18,  # Same as 1.8
    (1, 10): _extract_ordering_from_query_18  # Same as 1.8
}


def extract_ordering(query):
    version = django.VERSION[:2]
    return get_factory_for_version(_ORDERING_FACTORY, version)(query)



from .kinds import LOCK_KINDS


class MemcacheLock(object):

    @classmethod
    def acquire(cls, identifier, wait=True, steal_after_ms=None, kind=LOCK_KINDS.STRONG):
        raise NotImplementedError()

    def release(self):
        raise NotImplementedError()
